import errno
import subprocess
import sys
import webbrowser
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from typing import Any, List
from zipfile import ZipFile

from briefcase.console import Log

if sys.version_info >= (3, 11):  # pragma: no-cover-if-lt-py311
    import tomllib
else:  # pragma: no-cover-if-gte-py311
    import tomli as tomllib

import tomli_w

from briefcase.commands import (
    BuildCommand,
    CreateCommand,
    OpenCommand,
    PackageCommand,
    PublishCommand,
    RunCommand,
    UpdateCommand,
)
from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError, BriefcaseConfigError


class StaticWebMixin:
    output_format = "static"
    platform = "web"

    def project_path(self, app):
        return self.bundle_path(app) / "www"

    def binary_path(self, app):
        return self.bundle_path(app) / "www/index.html"

    def wheel_path(self, app):
        return self.project_path(app) / "static/wheels"

    def distribution_path(self, app):
        return self.dist_path / f"{app.formal_name}-{app.version}.web.zip"


class StaticWebCreateCommand(StaticWebMixin, CreateCommand):
    description = "Create and populate a static web project."

    def output_format_template_context(self, app: AppConfig):
        """Add style framework details to the app template."""
        return {
            "style_framework": getattr(app, "style_framework", "None"),
        }


class StaticWebUpdateCommand(StaticWebCreateCommand, UpdateCommand):
    description = "Update an existing static web project."


class StaticWebOpenCommand(StaticWebMixin, OpenCommand):
    description = "Open the folder containing an existing static web project."


class StaticWebBuildCommand(StaticWebMixin, BuildCommand):
    description = "Build a static web project."

    def _trim_file(self, path, sentinel):
        """Re-write a file to strip any content after a sentinel line.

        The file is stored in-memory, so it shouldn't be used on files with a *lot* of
        content before the sentinel.

        :param path: The path to the file to be trimmed
        :param sentinel: The content of the sentinel line. This will become the last
            line in the trimmed file.
        """
        content = []
        with path.open("r", encoding="utf-8") as f:
            for line in f:
                if line.rstrip("\n") == sentinel:
                    content.append(line)
                    break
                else:
                    content.append(line)

        with path.open("w", encoding="utf-8") as f:
            for line in content:
                f.write(line)

    def _process_wheel(self, wheelfile, css_file):
        """Process a wheel, extracting any content that needs to be compiled into the
        final project.

        :param wheelfile: The path to the wheel file to be processed.
        :param css_file: A file handle, opened for write/append, to which any extracted
            CSS content will be appended.
        """
        package = " ".join(wheelfile.name.split("-")[:2])
        with ZipFile(wheelfile) as wheel:
            for filename in wheel.namelist():
                path = Path(filename)
                # Any CSS file in a `static` folder is appended
                if (
                    len(path.parts) > 1
                    and path.parts[1] == "static"
                    and path.suffix == ".css"
                ):
                    self.logger.info(f"    Found {filename}")
                    css_file.write(
                        "\n/*******************************************************\n"
                    )
                    css_file.write(f" * {package}::{'/'.join(path.parts[2:])}\n")
                    css_file.write(
                        " *******************************************************/\n\n"
                    )
                    css_file.write(wheel.read(filename).decode("utf-8"))

    def build_app(self, app: AppConfig, **kwargs):
        """Build the static web deployment for the application.

        :param app: The application to build
        """
        self.logger.info("Building web project...", prefix=app.app_name)

        if self.wheel_path(app).exists():
            with self.input.wait_bar("Removing old wheels..."):
                self.tools.shutil.rmtree(self.wheel_path(app))

        self.wheel_path(app).mkdir(parents=True)

        with self.input.wait_bar("Building app wheel..."):
            try:
                self.tools.subprocess.run(
                    [
                        sys.executable,
                        "-u",
                        "-X",
                        "utf8",
                        "-m",
                        "wheel",
                        "pack",
                        self.app_path(app),
                        "--dest-dir",
                        self.wheel_path(app),
                    ],
                    check=True,
                    encoding="UTF-8",
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to build wheel for app {app.app_name!r}"
                ) from e

        with self.input.wait_bar("Installing wheels for requirements..."):
            try:
                self.tools.subprocess.run(
                    [
                        sys.executable,
                        "-u",
                        "-X",
                        "utf8",
                        "-m",
                        "pip",
                        "wheel",
                        "--wheel-dir",
                        self.wheel_path(app),
                        "-r",
                        self.bundle_path(app) / "requirements.txt",
                    ]
                    + (["-vv"] if self.logger.is_deep_debug else []),
                    check=True,
                    encoding="UTF-8",
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to install requirements for app {app.app_name!r}"
                ) from e

        with self.input.wait_bar("Writing Pyscript configuration file..."):
            with (self.project_path(app) / "pyscript.toml").open("wb") as f:
                config = {
                    "name": app.formal_name,
                    "description": app.description,
                    "version": app.version,
                    "splashscreen": {"autoclose": True},
                    "terminal": False,
                    # Ensure that we're using Unix path separators, as the content
                    # will be parsed by pyscript in the browser.
                    "packages": [
                        f'/{"/".join(wheel.relative_to(self.project_path(app)).parts)}'
                        for wheel in sorted(self.wheel_path(app).glob("*.whl"))
                    ],
                }
                # Parse any additional pyscript.toml content, and merge it into
                # the overall content
                try:
                    extra = tomllib.loads(app.extra_pyscript_toml_content)
                    config.update(extra)
                except tomllib.TOMLDecodeError as e:
                    raise BriefcaseConfigError(
                        f"Extra pyscript.toml content isn't valid TOML: {e}"
                    ) from e
                except AttributeError:
                    pass

                # Write the final configuration.
                tomli_w.dump(config, f)

        self.logger.info("Compile static web content from wheels")
        with self.input.wait_bar("Compiling static web content from wheels..."):
            # Trim previously compiled content out of briefcase.css
            briefcase_css_path = self.project_path(app) / "static/css/briefcase.css"
            self._trim_file(
                briefcase_css_path,
                sentinel=" ******************* Wheel contributed styles **********************/",
            )

            # Extract static resources from packaged wheels
            for wheelfile in sorted(self.wheel_path(app).glob("*.whl")):
                self.logger.info(f"  Processing {wheelfile.name}...")
                with briefcase_css_path.open("a", encoding="utf-8") as css_file:
                    self._process_wheel(wheelfile, css_file=css_file)

        return {}


class HTTPHandler(SimpleHTTPRequestHandler):
    """Convert any HTTP request into a path request on the static content folder."""

    server: "LocalHTTPServer"

    def translate_path(self, path):
        return str(self.server.base_path / path[1:])

    def end_headers(self):
        self.send_header("Cache-Control", "no-cache, no-store, must-revalidate")
        self.send_header("Pragma", "no-cache")
        self.send_header("Expires", "0")
        super().end_headers()

    # Strip out ANSI escape sequences from messages. This was added in security releases
    # before Python 3.12.
    _control_char_table = str.maketrans(
        {c: rf"\x{c:02x}" for c in [*range(0x20), *range(0x7F, 0xA0)]}
    )
    _control_char_table[ord("\\")] = r"\\"

    def log_message(self, format: str, *args: Any) -> None:
        message = (format % args).translate(self._control_char_table)
        self.server.logger.info(
            f"{self.address_string()} - - [{self.log_date_time_string()}] {message}"
        )


class LocalHTTPServer(ThreadingHTTPServer):
    """An HTTP server that serves local static content."""

    def __init__(
        self, base_path, host, port, RequestHandlerClass=HTTPHandler, *, logger: Log
    ):
        self.base_path = base_path
        self.logger = logger
        super().__init__((host, port), RequestHandlerClass)


class StaticWebRunCommand(StaticWebMixin, RunCommand):
    description = "Run a static web project."

    def add_options(self, parser):
        super().add_options(parser)
        parser.add_argument(
            "--host",
            default="localhost",
            help="The host on which to run the server (default: localhost)",
            required=False,
        )
        parser.add_argument(
            "-p",
            "--port",
            default=8080,
            type=int,
            help="The port on which to run the server (default: 8080)",
            required=False,
        )
        parser.add_argument(
            "--no-browser",
            action="store_false",
            dest="open_browser",
            help="Don't open a web browser on the newly opened server",
            required=False,
        )

    def run_app(
        self,
        app: AppConfig,
        test_mode: bool,
        passthrough: List[str],
        host,
        port,
        open_browser,
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param test_mode: Boolean; Is the app running in test mode?
        :param passthrough: The list of arguments to pass to the app
        :param host: The host on which to run the server
        :param port: The port on which to run the server
        :param open_browser: Should a browser be opened on the newly started server.
        """
        if test_mode:
            raise BriefcaseCommandError("Briefcase can't run web apps in test mode.")

        self.logger.info("Starting web server...", prefix=app.app_name)

        # At least for now, there's no easy way to pass arguments to a web app.
        if passthrough:
            self.logger.warning(f"Ignoring passthrough arguments: {passthrough}")

        httpd = None
        try:
            # Create a local HTTP server
            try:
                httpd = LocalHTTPServer(
                    self.project_path(app),
                    host=host,
                    port=port,
                    logger=self.logger,
                )
            except OSError as e:
                if e.errno in (errno.EADDRINUSE, errno.ENOSR):
                    self.logger.warning(
                        f"Using a system-allocated port since port {port} is already in use. "
                        "Use -p/--port to manually specify a port."
                    )
                    httpd = LocalHTTPServer(
                        self.project_path(app),
                        host=host,
                        port=0,
                        logger=self.logger,
                    )
                else:
                    raise

            # Extract the host and port from the server. This is needed
            # because specifying a port of 0 lets the server pick a port.
            host, port = httpd.socket.getsockname()
            url = f"http://{host}:{port}"

            self.logger.info(f"Web server open on {url}")
            # If requested, open a browser tab on the newly opened server.
            if open_browser:
                webbrowser.open_new_tab(url)

            self.logger.info(
                "Web server log output (type CTRL-C to stop log)...",
                prefix=app.app_name,
            )
            self.logger.info("=" * 75)

            # Run the server.
            httpd.serve_forever()
        except PermissionError as e:
            if port < 1024:
                raise BriefcaseCommandError(
                    "Unable to start web server; Permission denied. Try using a port > 1023."
                ) from e
            else:
                raise BriefcaseCommandError(
                    "Unable to start web server; Permission denied. Did you specify a valid host and port?"
                ) from e
        except OSError as e:
            if e.errno in (errno.EADDRNOTAVAIL, errno.ENOSTR):
                raise BriefcaseCommandError(
                    f"Unable to start web server. {host} is not a valid hostname."
                ) from e
            else:
                raise BriefcaseCommandError(f"Unable to start web server. {e}") from e
        except OverflowError as e:
            raise BriefcaseCommandError(
                "Unable to start web server. Port must be in the range 0-65535."
            ) from e
        except KeyboardInterrupt:
            # CTRL-C is the accepted way to stop the server.
            httpd.shutdown()
        finally:
            if httpd:
                with self.input.wait_bar("Shutting down server..."):
                    httpd.server_close()

            # Not sure why, but this is needed to mollify coverage for the
            # "test_cleanup_server_error" case. Without this pass, a missing branch
            # is reported for the "if httpd: -> exit" branch
            pass

        return {}


class StaticWebPackageCommand(StaticWebMixin, PackageCommand):
    description = "Package a static web app."

    @property
    def packaging_formats(self):
        return ["zip"]

    @property
    def default_packaging_format(self):
        return "zip"

    def package_app(self, app: AppConfig, **kwargs):
        """Package an app for distribution.

        :param app: The app to package.
        """
        self.logger.info(
            "Packaging web app for distribution...",
            prefix=app.app_name,
        )

        with self.input.wait_bar("Building archive..."):
            self.tools.shutil.make_archive(
                self.distribution_path(app).with_suffix(""),
                format="zip",
                root_dir=self.project_path(app),
            )


class StaticWebPublishCommand(StaticWebMixin, PublishCommand):
    description = "Publish a static web app."
    publication_channels = ["s3"]
    default_publication_channel = "s3"


# Declare the briefcase command bindings
create = StaticWebCreateCommand
update = StaticWebUpdateCommand
open = StaticWebOpenCommand
build = StaticWebBuildCommand
run = StaticWebRunCommand
package = StaticWebPackageCommand
publish = StaticWebPublishCommand



import subprocess
from pathlib import Path

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.xcode import get_identities


def security_result(name):
    """Load a security result file from the sample directory, and return the content."""
    filename = Path(__file__).parent / "security" / f"{name}.out"
    with filename.open(encoding="utf-8") as f:
        return f.read()


def test_security_missing(mock_tools):
    """If security is missing or fails to start, an exception is raised."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["security", "find-identities", "-v", "-p", "codesigning"], returncode=1
    )

    with pytest.raises(BriefcaseCommandError):
        get_identities(mock_tools, "codesigning")

    # The filter is passed through
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["security", "find-identity", "-v", "-p", "codesigning"]
    )


def test_invalid_profile(mock_tools):
    """If the requested profile is invalid, an exception is raised."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["security", "find-identities", "-v", "-p", "jabberwock"], returncode=2
    )

    with pytest.raises(BriefcaseCommandError):
        get_identities(mock_tools, "jabberwock")

    # The filter is passed through
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["security", "find-identity", "-v", "-p", "jabberwock"]
    )


def test_no_identities(mock_tools):
    """If there are no identities available, no simulators will be found."""
    mock_tools.subprocess.check_output.return_value = security_result("no-identities")

    simulators = get_identities(mock_tools, "codesigning")

    # The filter is passed through
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["security", "find-identity", "-v", "-p", "codesigning"]
    )

    assert simulators == {}


def test_one_identity(mock_tools):
    """If there is one identity available, it is returned."""
    mock_tools.subprocess.check_output.return_value = security_result("one-identity")

    simulators = get_identities(mock_tools, "codesigning")

    # The filter is passed through
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["security", "find-identity", "-v", "-p", "codesigning"]
    )

    assert simulators == {
        "38EBD6F8903EC63C238B04C1067833814CE47CA3": "Developer ID Application: Example Corporation Ltd (Z2K4383DLE)",
    }


def test_multiple_identities(mock_tools):
    """If there are multiple identities available, they are all returned."""
    mock_tools.subprocess.check_output.return_value = security_result(
        "multiple-identities"
    )

    simulators = get_identities(mock_tools, "codesigning")

    # The filter is passed through
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["security", "find-identity", "-v", "-p", "codesigning"]
    )

    assert simulators == {
        "38EBD6F8903EC63C238B04C1067833814CE47CA3": "Developer ID Application: Example Corporation Ltd (Z2K4383DLE)",
        "11E77FB58F13F6108B38110D5D92233C58ED38C5": "iPhone Developer: Jane Smith (BXAH5H869S)",
        "F8903EC63C238B04C1067833814CE47CA338EBD6": "Developer ID Application: Other Corporation Ltd (83DLZ2K43E)",
    }


def test_no_profile(mock_tools):
    """If no profile is specified, no filtering is done."""
    mock_tools.subprocess.check_output.return_value = security_result(
        "multiple-identities"
    )

    simulators = get_identities(mock_tools)

    # No filter is passed through
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["security", "find-identity", "-v"]
    )

    assert simulators == {
        "38EBD6F8903EC63C238B04C1067833814CE47CA3": "Developer ID Application: Example Corporation Ltd (Z2K4383DLE)",
        "11E77FB58F13F6108B38110D5D92233C58ED38C5": "iPhone Developer: Jane Smith (BXAH5H869S)",
        "F8903EC63C238B04C1067833814CE47CA338EBD6": "Developer ID Application: Other Corporation Ltd (83DLZ2K43E)",
    }



from unittest.mock import ANY, MagicMock


def test_default_and_intro_is_used(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    def get_source_dir_hint(*args, **kwargs):
        return "SOME_DIRECTORY", "SOME_DESCRIPTION"

    monkeypatch.setattr(convert_command, "get_source_dir_hint", get_source_dir_hint)

    convert_command.input_source_dir("app-name", "app_name", None)
    mock_input_text.assert_called_once_with(
        intro="SOME_DESCRIPTION",
        variable="source directory",
        default="SOME_DIRECTORY",
        validator=ANY,
        override_value=None,
    )


def test_default_and_intro_uses_override(convert_command, monkeypatch):
    def get_source_dir_hint(*args, **kwargs):
        return "SOME_DIRECTORY", "SOME_DESCRIPTION"

    monkeypatch.setattr(convert_command, "get_source_dir_hint", get_source_dir_hint)
    (convert_command.base_path / "OVERRIDE_VALUE" / "app_name").mkdir(parents=True)
    (
        convert_command.base_path / "OVERRIDE_VALUE" / "app_name" / "__main__.py"
    ).write_text("", encoding="utf-8")
    assert (
        convert_command.input_source_dir(
            "app-name", "app_name", "OVERRIDE_VALUE/app_name"
        )
        == "OVERRIDE_VALUE/app_name"
    )


def test_prompted_source_dir(convert_command):
    """You can type in the source dir."""
    (convert_command.base_path / "src/app_name").mkdir(parents=True)
    (convert_command.base_path / "src/app_name" / "__main__.py").write_text(
        "", encoding="utf-8"
    )
    convert_command.input.values = ["src/app_name"]

    assert (
        convert_command.input_source_dir("app-name", "app_name", None) == "src/app_name"
    )



import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import UnsupportedHostError
from briefcase.platforms.web.static import StaticWebCreateCommand


@pytest.fixture
def create_command(tmp_path):
    return StaticWebCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("host_os", ["WeirdOS"])
def test_unsupported_host_os(create_command, host_os):
    """Error raised for an unsupported OS."""
    create_command.tools.host_os = host_os

    with pytest.raises(UnsupportedHostError, match="This command is not supported on"):
        create_command()


def test_output_format_template_context_with_style_framework(
    create_command, first_app_config
):
    """If the app defines a style framework, it is included in the template context."""
    first_app_config.style_framework = "Souperstyler v1.2"

    assert create_command.output_format_template_context(first_app_config) == {
        "style_framework": "Souperstyler v1.2",
    }


def test_output_format_template_context_without_style_framework(
    create_command,
    first_app_config,
):
    """If the app doesn't define a style framework, the template context still has an
    entry."""
    assert create_command.output_format_template_context(first_app_config) == {
        "style_framework": "None",
    }



import subprocess

import pytest

from briefcase.console import LogLevel
from briefcase.exceptions import BriefcaseCommandError

from ...utils import create_file, file_content


@pytest.mark.parametrize("verbose", [True, False])
def test_thin_binary(dummy_command, verbose, tmp_path, capsys):
    """A thin binary is left as-is."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create a source binary.
    create_file(tmp_path / "path/to/file.dylib", "dylib-original")

    # Mock the result of the "lipo info" call.
    dummy_command.tools.subprocess.check_output.return_value = (
        "Non-fat file: path/to/file.dylib is architecture: gothic\n"
    )

    # Thin the binary; this is effectively a no-op
    dummy_command.ensure_thin_binary(
        tmp_path / "path/to/file.dylib",
        arch="gothic",
    )

    # Lipo -info was invoked
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "lipo",
            "-info",
            tmp_path / "path/to/file.dylib",
        ],
    )

    # Lipo -thin was *not* invoked
    dummy_command.tools.subprocess.run.assert_not_called()

    # The original file is unmodified.
    assert file_content(tmp_path / "path/to/file.dylib") == "dylib-original"

    # Output only happens if in debug mode
    output = capsys.readouterr().out.split("\n")
    assert len(output) == (2 if verbose else 1)


@pytest.mark.parametrize("verbose", [True, False])
def test_fat_dylib(dummy_command, verbose, tmp_path, capsys):
    """A fat binary library can be thinned."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create a source binary.
    create_file(tmp_path / "path/to/file.dylib", "dylib-fat")

    # Mock the result of the "lipo info" call.
    dummy_command.tools.subprocess.check_output.return_value = (
        "Architectures in the fat file: path/to/file.dylib are: modern gothic\n"
    )

    # Mock the result of successfully thinning a library
    def thin_dylib(*args, **kwargs):
        create_file(args[0][args[0].index("-output") + 1], "dylib-thin")

    dummy_command.tools.subprocess.run.side_effect = thin_dylib

    # Thin the binary to the "gothic" architecture
    dummy_command.ensure_thin_binary(
        tmp_path / "path/to/file.dylib",
        arch="gothic",
    )

    # Lipo -info was invoked
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "lipo",
            "-info",
            tmp_path / "path/to/file.dylib",
        ],
    )

    # Lipo -thin was invoked
    dummy_command.tools.subprocess.run.assert_called_once_with(
        [
            "lipo",
            "-thin",
            "gothic",
            "-output",
            tmp_path / "path/to/file.dylib.gothic",
            tmp_path / "path/to/file.dylib",
        ],
        check=True,
    )

    # The original file now has the thinned content.
    assert file_content(tmp_path / "path/to/file.dylib") == "dylib-thin"

    # Output only happens if in debug mode
    output = capsys.readouterr().out.split("\n")
    assert len(output) == (2 if verbose else 1)


@pytest.mark.parametrize("verbose", [True, False])
def test_fat_dylib_arch_mismatch(dummy_command, verbose, tmp_path, capsys):
    """If a fat binary doesn't contain the target architecture, an error is raised."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create a source binary.
    create_file(tmp_path / "path/to/file.dylib", "dylib-fat")

    # Mock the result of the "lipo info" call.
    dummy_command.tools.subprocess.check_output.return_value = (
        "Architectures in the fat file: path/to/file.dylib are: modern artdeco\n"
    )

    # Thin the binary to the "gothic" architecture. This will raise an exception
    with pytest.raises(
        BriefcaseCommandError,
        match=r"file\.dylib does not contain a gothic slice",
    ):
        dummy_command.ensure_thin_binary(
            tmp_path / "path/to/file.dylib",
            arch="gothic",
        )

    # Lipo -info was invoked
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "lipo",
            "-info",
            tmp_path / "path/to/file.dylib",
        ],
    )

    # Lipo -thin was *not* invoked
    dummy_command.tools.subprocess.run.assert_not_called()


@pytest.mark.parametrize("verbose", [True, False])
def test_fat_dylib_unknown_info(dummy_command, verbose, tmp_path, capsys):
    """If the lipo info call succeeds, but generates unknown output, an error is
    raised."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create a source binary.
    create_file(tmp_path / "path/to/file.dylib", "dylib-fat")

    # Mock the result of the "lipo info" call.
    dummy_command.tools.subprocess.check_output.return_value = (
        "This is unexpected output...\n"
    )

    # Thin the binary to the "gothic" architecture. This will raise an exception
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to determine architectures in .*file\.dylib",
    ):
        dummy_command.ensure_thin_binary(
            tmp_path / "path/to/file.dylib",
            arch="gothic",
        )

    # Lipo -info was invoked
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "lipo",
            "-info",
            tmp_path / "path/to/file.dylib",
        ],
    )

    # Lipo -thin was *not* invoked
    dummy_command.tools.subprocess.run.assert_not_called()


def test_lipo_info_fail(dummy_command, tmp_path):
    """If lipo can't inspect a binary, an error is raised."""
    # Create a source binary.
    create_file(tmp_path / "path/to/file.dylib", "dylib-fat")

    # Mock the result of the "lipo info" call.
    dummy_command.tools.subprocess.check_output.side_effect = (
        subprocess.CalledProcessError(cmd="lipo -info", returncode=-1)
    )

    # Thin the binary to the "gothic" architecture. This will raise an exception
    with pytest.raises(
        BriefcaseCommandError, match=r"Unable to inspect architectures in .*file\.dylib"
    ):
        dummy_command.ensure_thin_binary(
            tmp_path / "path/to/file.dylib",
            arch="gothic",
        )

    # Lipo -info was invoked
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "lipo",
            "-info",
            tmp_path / "path/to/file.dylib",
        ],
    )

    # Lipo -thin was not invoked
    dummy_command.tools.subprocess.run.assert_not_called()


@pytest.mark.parametrize("verbose", [True, False])
def test_lipo_thin_fail(dummy_command, verbose, tmp_path, capsys):
    """If lipo fails thinning the binary, an error is raised."""
    if verbose:
        dummy_command.logger.verbosity = LogLevel.VERBOSE

    # Create a source binary.
    create_file(tmp_path / "path/to/file.dylib", "dylib-fat")

    # Mock the result of the "lipo -info" call.
    dummy_command.tools.subprocess.check_output.return_value = (
        "Architectures in the fat file: path/to/file.dylib are: modern gothic\n"
    )

    # Mock the result of the failed "lipo -thin" call.
    dummy_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd="lipo -thin", returncode=-1
    )

    # Thin the binary to the "gothic" architecture. This will raise an exception
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to create thin binary from .*file.dylib",
    ):
        dummy_command.ensure_thin_binary(
            tmp_path / "path/to/file.dylib",
            arch="gothic",
        )

    # Lipo -info was invoked
    dummy_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "lipo",
            "-info",
            tmp_path / "path/to/file.dylib",
        ],
    )

    # Lipo -thin was invoked
    dummy_command.tools.subprocess.run.assert_called_once_with(
        [
            "lipo",
            "-thin",
            "gothic",
            "-output",
            tmp_path / "path/to/file.dylib.gothic",
            tmp_path / "path/to/file.dylib",
        ],
        check=True,
    )

    # The original file is unmodified.
    assert file_content(tmp_path / "path/to/file.dylib") == "dylib-fat"

    # Output only happens if in debug mode
    output = capsys.readouterr().out.split("\n")
    assert len(output) == (2 if verbose else 1)



import subprocess
from unittest import mock

import pytest

from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.integrations.flatpak import Flatpak


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.flatpak = "tool"

    tool = Flatpak.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.flatpak


@pytest.mark.parametrize("host_os", ["Darwin", "Windows", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{Flatpak.name} is not supported on {host_os}",
    ):
        Flatpak.verify(mock_tools)


def test_flatpak_not_installed(mock_tools):
    """If flatpak isn't installed, an error is raised."""
    # Mock the response from a missing tool
    mock_tools.subprocess.check_output.side_effect = FileNotFoundError

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase requires the Flatpak toolchain, but it does not appear to be installed.",
    ):
        Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
        ],
        any_order=False,
    )


def test_flatpak_error(mock_tools):
    """If `flatpak --version` fails, an error is raised."""
    # Mock the side effect of an unsuccessful call to flatpak
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd="flatpak --version", returncode=1
    )

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to invoke flatpak.",
    ):
        Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
        ],
        any_order=False,
    )


def test_flatpak_old(mock_tools):
    """If `flatpak --version` returns an old version, an error is raised."""
    # Mock the side effect of a successful call to an older version of Flatpak
    mock_tools.subprocess.check_output.return_value = "Flatpak 0.10.9\n"

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase requires Flatpak 1.0 or later.",
    ):
        Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
        ],
        any_order=False,
    )


def test_flatpak_builder_not_installed(mock_tools):
    """If flatpak isn't installed, an error is raised."""
    # Mock the side effect of a call to a missing flatpak-builder
    mock_tools.subprocess.check_output.side_effect = [
        "Flatpak 1.12.7",
        FileNotFoundError,
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase requires the full Flatpak development toolchain, but flatpak-builder",
    ):
        Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
            mock.call(["flatpak-builder", "--version"]),
        ],
        any_order=False,
    )


def test_flatpak_builder_error(mock_tools):
    """If `flatpak-builder --version` fails, an error is raised."""
    # Mock the side effect of an unsuccessful call to flatpak-builder
    mock_tools.subprocess.check_output.side_effect = [
        "Flatpak 1.12.7",
        subprocess.CalledProcessError(cmd="flatpak-builder --version", returncode=1),
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to invoke flatpak-builder.",
    ):
        Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
            mock.call(["flatpak-builder", "--version"]),
        ],
        any_order=False,
    )


def test_flatpak_builder_old(mock_tools):
    """If the version of flatpak-builder is old, an error is raised."""
    # Mock the side effect of a call to a missing flatpak-builder
    mock_tools.subprocess.check_output.side_effect = [
        "Flatpak 1.12.7",
        "flatpak-builder 0.10.9",
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase requires flatpak-builder 1.0 or later.",
    ):
        Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
            mock.call(["flatpak-builder", "--version"]),
        ],
        any_order=False,
    )


@pytest.mark.parametrize(
    "flatpak_builder_version",
    [
        # Ubuntu 22.04; flatpak-builder < 1.3
        "flatpak-builder 1.2.2",
        # Fedora 38; flatpak-builder >= 1.3
        "flatpak-builder-1.3.3",
    ],
)
def test_installed(mock_tools, flatpak_builder_version, capsys):
    """If flatpak is installed, it can be verified."""
    mock_tools.subprocess.check_output.side_effect = [
        "Flatpak 1.12.7",
        flatpak_builder_version,
    ]

    Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
            mock.call(["flatpak-builder", "--version"]),
        ],
        any_order=False,
    )

    # We shouldn't get any warnings about unknown versions.
    output = capsys.readouterr()
    assert (
        "** WARNING: Unable to determine the version of flatpak-builder"
        not in output.out
    )
    assert output.err == ""


@pytest.mark.parametrize(
    "flatpak_version",
    [
        # Different tool identification
        "not-flatpak 1.2.2",
        # No version ID
        "Flatpak",
        # Non-integer version ID
        "Flatpak x.y.z",
    ],
)
def test_installed_unknown_flatpak_version(mock_tools, flatpak_version, capsys):
    """If flatpak is installed, but the version can't be determined, a warning is
    raised, but flatpak is verified."""
    mock_tools.subprocess.check_output.side_effect = [
        flatpak_version,
        "flatpak-builder 1.2.2",
    ]

    Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
            mock.call(["flatpak-builder", "--version"]),
        ],
        any_order=False,
    )

    output = capsys.readouterr()
    assert "** WARNING: Unable to determine the version of Flatpak" in output.out
    assert output.err == ""


@pytest.mark.parametrize(
    "builder_version",
    [
        # Different tool identification
        "not-flatpak-builder 1.2.2",
        # No version ID
        "flatpak-builder",
        # Non-integer version ID
        "flatpak-builder x.y.z",
    ],
)
def test_installed_unknown_builder_version(mock_tools, builder_version, capsys):
    """If flatpak-builder is installed, but the version can't be determined, a warning
    is raised, but flatpak is verified."""
    mock_tools.subprocess.check_output.side_effect = [
        "Flatpak 1.12.7",
        builder_version,
    ]

    Flatpak.verify(mock_tools)

    mock_tools.subprocess.check_output.assert_has_calls(
        [
            mock.call(["flatpak", "--version"]),
            mock.call(["flatpak-builder", "--version"]),
        ],
        any_order=False,
    )

    output = capsys.readouterr()
    assert (
        "** WARNING: Unable to determine the version of flatpak-builder" in output.out
    )
    assert output.err == ""



import json
import os
import subprocess
from pathlib import Path
from unittest.mock import call

import pytest

from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.integrations.visualstudio import VisualStudio

MSBUILD_OUTPUT = """Microsoft (R) Build Engine version 17.2.1+52cd2da31 for .NET Framework
Copyright (C) Microsoft Corporation. All rights reserved.

17.2.1.25201
"""


@pytest.fixture
def custom_msbuild_path(tmp_path):
    """Create a dummy MSBuild executable at a custom location."""
    msbuild_path = tmp_path / "custom/MSBuild.exe"
    msbuild_path.parent.mkdir(parents=True)
    with msbuild_path.open("w", encoding="utf-8") as f:
        f.write("Dummy MSBuild")

    return msbuild_path


@pytest.fixture
def vswhere_path(tmp_path):
    """Create a dummy vswhere executable."""
    vswhere_path = (
        tmp_path
        / "Program Files (x86)"
        / "Microsoft Visual Studio"
        / "Installer"
        / "vswhere.exe"
    )

    vswhere_path.parent.mkdir(parents=True)
    with vswhere_path.open("w", encoding="utf-8") as f:
        f.write("Dummy vswhere")

    return vswhere_path


@pytest.fixture
def msbuild_path(tmp_path):
    """Create a dummy MSBuild executable."""
    msbuild_path = tmp_path / "Visual Studio/MSBuild/Current/Bin/MSBuild.exe"
    msbuild_path.parent.mkdir(parents=True)
    with msbuild_path.open("w", encoding="utf-8") as f:
        f.write("Dummy MSBuild")

    return msbuild_path


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.visualstudio = "tool"

    tool = VisualStudio.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.visualstudio


@pytest.mark.parametrize("host_os", ["Darwin", "Linux", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{VisualStudio.name} is not supported on {host_os}",
    ):
        VisualStudio.verify(mock_tools)


def test_msbuild_on_path(mock_tools):
    """If MSBuild is on the path, that version is used."""
    # MSBuild is on the path, so check_output succeeds
    mock_tools.subprocess.check_output.return_value = MSBUILD_OUTPUT

    # Verify the installation
    visualstudio = VisualStudio.verify(mock_tools)

    # Visual studio is configured to use an MSBuild with no path,
    # which provides no metadata.
    assert visualstudio.msbuild_path == Path("MSBuild.exe")
    assert visualstudio.install_metadata is None

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [call(["MSBuild.exe", "--version"])],
        any_order=False,
    )


def test_msbuild_on_path_corrupt(mock_tools):
    """If MSBuild is on the path, but it cannot be invoked, an error is raised."""
    # MSBuild is on the path, but raises an error when invoked
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=1,
        cmd="MSBuild.exe",
    )

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"MSBuild is on the path, but Briefcase cannot start it.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [call(["MSBuild.exe", "--version"])],
        any_order=False,
    )


def test_msbuild_envvar(mock_tools, custom_msbuild_path):
    """If MSBUILD is set in the environment, that executable is used."""
    # Point at the dummy MSBuild executable
    mock_tools.os.environ["MSBUILD"] = custom_msbuild_path

    # MSBuild is not on the path, but the
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # Not on path
        0,  # Custom location succeeds
    ]

    # Verify the installation
    visualstudio = VisualStudio.verify(mock_tools)

    # Visual studio is configured to use an MSBuild at the specified path
    # which provides no metadata.
    assert visualstudio.msbuild_path == custom_msbuild_path
    assert visualstudio.install_metadata is None

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call([custom_msbuild_path, "--version"]),
        ],
        any_order=False,
    )


def test_msbuild_envvar_doesnt_exist(mock_tools, tmp_path):
    """If MSBUILD is set in the environment, but it points to a non-existent file, an
    error is raised."""
    # Point at an MSBuild that does not exist
    mock_tools.os.environ["MSBUILD"] = tmp_path / "custom/MSBuild.exe"

    # MSBuild is not on the path
    mock_tools.subprocess.check_output.side_effect = FileNotFoundError

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"The location referenced by the environment variable MSBUILD:",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
        ],
        any_order=False,
    )


def test_msbuild_envvar_bad_executable(mock_tools, custom_msbuild_path):
    """If MSBUILD is set in the environment, but it can't be invoked, an error is
    raised."""
    # Point at the dummy MSBuild executable
    mock_tools.os.environ["MSBUILD"] = custom_msbuild_path

    # MSBuild is not on the path, and can't be invoked at the custom location
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # Not on path
        subprocess.CalledProcessError(-1, custom_msbuild_path),  # Custom location fails
    ]

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"MSBuild appears to exist, but Briefcase can't start it.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call([custom_msbuild_path, "--version"]),
        ],
        any_order=False,
    )


def test_vswhere_does_not_exist(mock_tools):
    """If VSWhere does not exist, an error is raised."""
    # MSBuild is not on the path
    mock_tools.subprocess.check_output.side_effect = FileNotFoundError

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Visual Studio does not appear to be installed.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
        ],
        any_order=False,
    )


def test_programfiles_envvar_does_not_exist(mock_tools):
    """If ProgramFiles(x86) env var is not set, an error is raised."""
    # MSBuild is not on the path
    mock_tools.subprocess.check_output.side_effect = FileNotFoundError

    # ProgramFiles(x86) env var is not set
    mock_tools.os.environ.pop("ProgramFiles(x86)", None)

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Visual Studio does not appear to be installed.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
        ],
        any_order=False,
    )


def test_vswhere_bad_executable(mock_tools, vswhere_path):
    """If VSWhere exists, but cannot be executed, an error is raised."""
    # MSBuild is not on the path, and vswhere raises an error
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # MSBuild not on path
        subprocess.CalledProcessError(returncode=1, cmd=vswhere_path),  # vswhere fails
    ]

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Visual Studio appears to exist, but Briefcase can't retrieve installation metadata.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call(
                [vswhere_path, "-latest", "-prerelease", "-format", "json"],
                stderr=subprocess.STDOUT,
            ),
        ],
        any_order=False,
    )


def test_vswhere_bad_content(mock_tools, vswhere_path):
    """If VSWhere can be executed, but returns garbage content, an error is raised."""
    # MSBuild is not on the path, and vswhere returns non-JSON content
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # MSBuild not on path
        "This is not JSON content",  # vswhere returns non-JSON content
    ]

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Visual Studio appears to exist, but Briefcase can't retrieve installation metadata.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call(
                [vswhere_path, "-latest", "-prerelease", "-format", "json"],
                stderr=subprocess.STDOUT,
            ),
        ],
        any_order=False,
    )


def test_vswhere_non_list_content(mock_tools, vswhere_path):
    """If VSWhere can be executed, but the outermost content isn't a list, an error is
    raised."""
    # MSBuild is not on the path, and vswhere returns JSON content, but not in the format expected
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # MSBuild not on path
        '{"problem": "JSON but not a list"}',  # vswhere returns JSON content, but not as a list.
    ]

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Visual Studio appears to exist, but Briefcase can't retrieve installation metadata.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call(
                [vswhere_path, "-latest", "-prerelease", "-format", "json"],
                stderr=subprocess.STDOUT,
            ),
        ],
        any_order=False,
    )


def test_vswhere_empty_list_content(mock_tools, vswhere_path):
    """If VSWhere can be executed, but the outermost content is an empty list, an error
    is raised."""
    # MSBuild is not on the path, and vswhere returns JSON content, but not in the format expected
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # MSBuild not on path
        "[]",  # vswhere returns empty list JSON content
    ]

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Visual Studio appears to exist, but Briefcase can't retrieve installation metadata.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call(
                [vswhere_path, "-latest", "-prerelease", "-format", "json"],
                stderr=subprocess.STDOUT,
            ),
        ],
        any_order=False,
    )


def test_vswhere_msbuild_not_installed(mock_tools, tmp_path, vswhere_path):
    """If VSWhere can be executed, but it doesn't point at an MSBuild executable, an
    error is raised."""
    # MSBuild is not on the path; vswhere a valid location, but there's no MSBuild there.
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # MSBuild not on path
        json.dumps(
            [
                {
                    "instanceId": "deadbeef",
                    "installDate": "2022-07-14T10:42:37Z",
                    "installationPath": os.fsdecode(tmp_path / "Visual Studio"),
                }
            ]
        ),  # vswhere returns JSON content
    ]

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Your Visual Studio installation does not appear to provide MSBuild.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call(
                [vswhere_path, "-latest", "-prerelease", "-format", "json"],
                stderr=subprocess.STDOUT,
            ),
        ],
        any_order=False,
    )


def test_vswhere_msbuild_bad_executable(
    mock_tools,
    tmp_path,
    vswhere_path,
    msbuild_path,
):
    """If VSWhere points at an MSBuild executable, but that exe can't be started, an
    error is raised."""
    # MSBuild is not on the path; vswhere a valid location, but MSBuild fails.
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # MSBuild not on path
        json.dumps(
            [
                {
                    "instanceId": "deadbeef",
                    "installDate": "2022-07-14T10:42:37Z",
                    "installationPath": os.fsdecode(tmp_path / "Visual Studio"),
                }
            ]
        ),  # vswhere returns JSON content
        subprocess.CalledProcessError(
            returncode=1,
            cmd="MSBuild.exe",
        ),  # MSBuild fails
    ]

    # Verify the installation
    with pytest.raises(
        BriefcaseCommandError,
        match=r"MSBuild appears to exist, but Briefcase can't start it.",
    ):
        VisualStudio.verify(mock_tools)

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call(
                [vswhere_path, "-latest", "-prerelease", "-format", "json"],
                stderr=subprocess.STDOUT,
            ),
            call([msbuild_path, "--version"]),
        ],
        any_order=False,
    )


def test_vswhere_install(mock_tools, tmp_path, vswhere_path, msbuild_path):
    """If VSWhere points at a valid MSBuild executable, that executable is used."""
    # MSBuild is not on the path; vswhere a valid location, and MSBuild succeeds.
    mock_tools.subprocess.check_output.side_effect = [
        FileNotFoundError,  # MSBuild not on path
        json.dumps(
            [
                {
                    "instanceId": "deadbeef",
                    "installDate": "2022-07-14T10:42:37Z",
                    "installationPath": os.fsdecode(tmp_path / "Visual Studio"),
                }
            ]
        ),  # vswhere returns JSON content
        MSBUILD_OUTPUT,  # MSBuild Succeeds
    ]

    # Verify the installation
    visualstudio = VisualStudio.verify(mock_tools)

    assert visualstudio.msbuild_path == msbuild_path
    assert visualstudio.install_metadata["instanceId"] == "deadbeef"

    # Verification calls are as expected
    mock_tools.subprocess.check_output.assert_has_calls(
        [
            call(["MSBuild.exe", "--version"]),
            call(
                [vswhere_path, "-latest", "-prerelease", "-format", "json"],
                stderr=subprocess.STDOUT,
            ),
            call([msbuild_path, "--version"]),
        ],
        any_order=False,
    )



import sys

if sys.version_info >= (3, 10):  # pragma: no-cover-if-lt-py310
    from importlib.metadata import entry_points
else:  # pragma: no-cover-if-gte-py310
    # Before Python 3.10, entry_points did not support the group argument;
    # so, the backport package must be used on older versions.
    from importlib_metadata import entry_points


def get_platforms():
    return {
        entry_point.name: entry_point.load()
        for entry_point in entry_points(group="briefcase.platforms")
    }


def get_output_formats(platform):
    # The values for output format entry points are the importable module names,
    # which may not match the human-readable name. (e.g., the Xcode format is in
    # the briefcase.platforms.macOS.xcode module). The human-readable names for
    # formats must be extracted from the `output_format` attribute of a command
    # *in* the module; every output format has a `create` command, so we use
    # that.
    output_formats = {}
    for entry_point in entry_points(group=f"briefcase.formats.{platform}"):
        module = entry_point.load()
        output_formats[module.create.output_format] = module

    return output_formats



import shlex
import sys
from unittest import mock

import pytest

from briefcase import __version__, cmdline
from briefcase.commands import ConvertCommand, DevCommand, NewCommand, UpgradeCommand
from briefcase.console import Console, Log, LogLevel
from briefcase.exceptions import (
    InvalidFormatError,
    NoCommandError,
    UnsupportedCommandError,
)
from briefcase.platforms.linux.system import LinuxSystemCreateCommand
from briefcase.platforms.macOS.app import (
    macOSAppCreateCommand,
    macOSAppPublishCommand,
    macOSAppRunCommand,
)
from briefcase.platforms.windows.app import WindowsAppCreateCommand


@pytest.fixture
def logger() -> Log:
    return Log()


@pytest.fixture
def console() -> Console:
    return Console()


def do_cmdline_parse(args: list, logger: Log, console: Console):
    """Simulate process to parse command line."""
    Command, extra_cmdline = cmdline.parse_cmdline(args)
    cmd = Command(logger=logger, console=console)
    options, overrides = cmd.parse_options(extra=extra_cmdline)
    return cmd, options, overrides


def test_empty():
    """``briefcase`` returns basic usage."""
    with pytest.raises(NoCommandError, match=r"usage: briefcase") as excinfo:
        cmdline.parse_cmdline("".split())

    assert excinfo.value.msg.startswith(
        "usage: briefcase [-h] <command> [<platform>] [<format>] ...\n"
        "\n"
        "Briefcase is a tool for converting a Python project into a standalone native\n"
        "application for distribution.\n"
        "\n"
        "Commands:\n"
    )


def test_help_only():
    """``briefcase -h`` returns basic usage."""
    with pytest.raises(NoCommandError, match=r"usage: briefcase") as excinfo:
        cmdline.parse_cmdline("-h".split())

    assert excinfo.value.msg.startswith(
        "usage: briefcase [-h] <command> [<platform>] [<format>] ...\n"
        "\n"
        "Briefcase is a tool for converting a Python project into a standalone native\n"
        "application for distribution.\n"
        "\n"
        "Commands:\n"
    )


def test_version_only(capsys):
    """``briefcase -V`` returns current version."""
    with pytest.raises(SystemExit) as excinfo:
        cmdline.parse_cmdline("-V".split())

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Version is displayed.
    output = capsys.readouterr().out
    assert output == f"{__version__}\n"


def test_unknown_command():
    """``briefcase foobar`` fails as an invalid command."""
    with pytest.raises(SystemExit) as excinfo:
        cmdline.parse_cmdline("foobar".split())

    assert excinfo.value.code == 2
    assert excinfo.value.__context__.argument_name == "command"
    assert excinfo.value.__context__.message.startswith(
        "invalid choice: 'foobar' (choose from"
    )


@pytest.mark.parametrize(
    "cmdline, expected_options, expected_overrides",
    [
        (
            "new",
            {
                "template": None,
                "template_branch": None,
                "project_overrides": None,
            },
            {},
        ),
        (
            "new --template=path/to/template --template-branch=experiment -C version=\\'1.2.3\\' -C other=42",
            {
                "template": "path/to/template",
                "template_branch": "experiment",
                "project_overrides": None,
            },
            {
                "version": "1.2.3",
                "other": 42,
            },
        ),
    ],
)
def test_new_command(logger, console, cmdline, expected_options, expected_overrides):
    """``briefcase new`` returns the New command."""
    cmd, options, overrides = do_cmdline_parse(shlex.split(cmdline), logger, console)

    assert isinstance(cmd, NewCommand)
    assert cmd.platform == "all"
    assert cmd.output_format is None
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == expected_options
    assert overrides == expected_overrides


@pytest.mark.parametrize(
    "cmdline, expected_options, expected_overrides",
    [
        (
            "convert",
            {
                "template": None,
                "template_branch": None,
                "project_overrides": None,
            },
            {},
        ),
        (
            "convert --template=path/to/template --template-branch=experiment -C version=\\'1.2.3\\' -C other=42",
            {
                "template": "path/to/template",
                "template_branch": "experiment",
                "project_overrides": None,
            },
            {
                "version": "1.2.3",
                "other": 42,
            },
        ),
    ],
)
def test_convert_command(
    logger, console, cmdline, expected_options, expected_overrides
):
    """``briefcase convert`` returns the Convert command."""
    cmd, options, overrides = do_cmdline_parse(shlex.split(cmdline), logger, console)

    assert isinstance(cmd, ConvertCommand)
    assert cmd.platform == "all"
    assert cmd.output_format is None
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == expected_options
    assert overrides == expected_overrides


# Common tests for dev and run commands.
def dev_run_parameters(command):
    return [
        (f"{command} {args}", expected, overrides)
        for args, expected, overrides in [
            ("", {}, {}),
            ("-r", {"update_requirements": True}, {}),
            (
                "-r -C version=\\'1.2.3\\' -C other=42",
                {"update_requirements": True},
                {
                    "version": "1.2.3",
                    "other": 42,
                },
            ),
            ("--update-requirements", {"update_requirements": True}, {}),
            ("--test", {"test_mode": True}, {}),
            ("--test -r", {"test_mode": True, "update_requirements": True}, {}),
            ("--", {}, {}),
            ("-- ''", {"passthrough": [""]}, {}),
            ("-- --test", {"passthrough": ["--test"]}, {}),
            ("--test -- --test", {"test_mode": True, "passthrough": ["--test"]}, {}),
            ("--test -- -r", {"test_mode": True, "passthrough": ["-r"]}, {}),
            (
                "-r -- --test",
                {"update_requirements": True, "passthrough": ["--test"]},
                {},
            ),
            ("-- -y --no maybe", {"passthrough": ["-y", "--no", "maybe"]}, {}),
            (
                "--test -- -y --no maybe",
                {"test_mode": True, "passthrough": ["-y", "--no", "maybe"]},
                {},
            ),
        ]
    ]


@pytest.mark.parametrize(
    "cmdline, expected_options, expected_overrides",
    dev_run_parameters("dev")
    + [
        ("dev --no-run", {"run_app": False}, {}),
    ],
)
def test_dev_command(
    monkeypatch,
    logger,
    console,
    cmdline,
    expected_options,
    expected_overrides,
):
    """``briefcase dev`` returns the Dev command."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    cmd, options, overrides = do_cmdline_parse(shlex.split(cmdline), logger, console)

    assert isinstance(cmd, DevCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format is None
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {
        "appname": None,
        "update_requirements": False,
        "run_app": True,
        "test_mode": False,
        "passthrough": [],
        **expected_options,
    }
    assert overrides == expected_overrides


@pytest.mark.parametrize(
    "cmdline, expected_options, expected_overrides",
    dev_run_parameters("run")
    + [
        ("run -u", {"update": True}, {}),
        ("run --update", {"update": True}, {}),
        ("run --update-resources", {"update_resources": True}, {}),
        ("run --update-support", {"update_support": True}, {}),
        ("run --update-stub", {"update_stub": True}, {}),
        ("run --no-update", {"no_update": True}, {}),
    ],
)
def test_run_command(
    monkeypatch,
    logger,
    console,
    cmdline,
    expected_options,
    expected_overrides,
):
    """``briefcase run`` returns the Run command for the correct platform."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    cmd, options, overrides = do_cmdline_parse(shlex.split(cmdline), logger, console)

    assert isinstance(cmd, macOSAppRunCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {
        "appname": None,
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "passthrough": [],
        **expected_options,
    }
    assert overrides == expected_overrides


@pytest.mark.parametrize(
    "cmdline,expected_options,expected_overrides",
    [
        (
            "upgrade",
            {
                "list_tools": False,
                "tool_list": [],
            },
            {},
        ),
        (
            "upgrade -C version='1.2.3' -C other=42",
            {
                "list_tools": False,
                "tool_list": [],
            },
            {
                "version": "1.2.3",
                "other": 42,
            },
        ),
    ],
)
def test_upgrade_command(
    monkeypatch,
    logger,
    console,
    cmdline,
    expected_options,
    expected_overrides,
):
    """``briefcase upgrade`` returns the upgrade command."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    cmd, options, overrides = do_cmdline_parse(cmdline.split(), logger, console)

    assert isinstance(cmd, UpgradeCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format is None
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == expected_options
    assert overrides == expected_overrides


def test_bare_command(monkeypatch, logger, console):
    """``briefcase create`` returns the macOS create app command."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    cmd, options, overrides = do_cmdline_parse("create".split(), logger, console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}
    assert overrides == {}


@pytest.mark.skipif(sys.platform != "linux", reason="requires Linux")
def test_linux_default(logger, console):
    """``briefcase create`` returns the linux create system command on Linux."""

    cmd, options, overrides = do_cmdline_parse("create".split(), logger, console)

    assert isinstance(cmd, LinuxSystemCreateCommand)
    assert cmd.platform == "linux"
    assert cmd.output_format == "system"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}


@pytest.mark.skipif(sys.platform != "darwin", reason="requires macOS")
def test_macOS_default(logger, console):
    """``briefcase create`` returns the macOS create command on Linux."""

    cmd, options, overrides = do_cmdline_parse("create".split(), logger, console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}
    assert overrides == {}


@pytest.mark.skipif(sys.platform != "win32", reason="requires Windows")
def test_windows_default(logger, console):
    """``briefcase create`` returns the Windows create app command on Windows."""

    cmd, options, overrides = do_cmdline_parse("create".split(), logger, console)

    assert isinstance(cmd, WindowsAppCreateCommand)
    assert cmd.platform == "windows"
    assert cmd.output_format == "app"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}
    assert overrides == {}


def test_bare_command_help(monkeypatch, capsys, logger, console):
    """``briefcase create -h`` returns the macOS create app command help."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create -h".split(), logger, console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Help message is for default platform and format
    output = capsys.readouterr().out
    assert output.startswith(
        "usage: briefcase create macOS app [-h] [-C KEY=VALUE] [-v] [-V] [--no-input]\n"
        "                                  [--log]\n"
        "\n"
        "Create and populate a macOS app.\n"
    )


def test_bare_command_version(capsys, logger, console):
    """``briefcase create -V`` returns the version."""
    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create -V".split(), logger, console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Version is displayed.
    output = capsys.readouterr().out
    assert output == f"{__version__}\n"


def test_command_unknown_platform(monkeypatch, logger, console):
    """``briefcase create foobar`` raises an unknown platform error."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create foobar".split(), logger, console)

    assert excinfo.value.code == 2
    assert excinfo.value.__context__.argument_name == "platform"
    assert excinfo.value.__context__.message.startswith(
        "invalid choice: 'foobar' (choose from"
    )


def test_command_explicit_platform(monkeypatch, logger, console):
    """``briefcase create linux`` returns linux create app command."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    cmd, options, overrides = do_cmdline_parse("create linux".split(), logger, console)

    assert isinstance(cmd, LinuxSystemCreateCommand)
    assert cmd.platform == "linux"
    assert cmd.output_format == "system"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}
    assert overrides == {}


def test_command_explicit_platform_case_handling(monkeypatch, logger, console):
    """``briefcase create macOS`` returns macOS create app command."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    # This is all lower case; the command normalizes to macOS
    cmd, options, overrides = do_cmdline_parse("create macOS".split(), logger, console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}
    assert overrides == {}


def test_command_explicit_platform_help(monkeypatch, capsys, logger, console):
    """``briefcase create macOS -h`` returns the macOS create app command help."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create macOS -h".split(), logger, console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Help message is for default platform and format
    output = capsys.readouterr().out
    assert output.startswith(
        "usage: briefcase create macOS app [-h] [-C KEY=VALUE] [-v] [-V] [--no-input]\n"
        "                                  [--log]\n"
        "\n"
        "Create and populate a macOS app.\n"
    )


def test_command_explicit_format(monkeypatch, logger, console):
    """``briefcase create macOS app`` returns the macOS create app command."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    cmd, options, overrides = do_cmdline_parse(
        "create macOS app".split(), logger, console
    )

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}
    assert overrides == {}


def test_command_unknown_format(monkeypatch, logger, console):
    """``briefcase create macOS foobar`` returns an invalid format error."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    expected_exc_regex = r"Invalid format 'foobar'; \(choose from: app, Xcode\)"
    with pytest.raises(InvalidFormatError, match=expected_exc_regex):
        do_cmdline_parse("create macOS foobar".split(), logger, console)


def test_command_explicit_unsupported_format(monkeypatch, logger, console):
    """``briefcase create macOS homebrew`` raises an error because the format isn't
    supported (yet)"""
    # Mock the output formats to include a "homebrew" backend with no commands.
    monkeypatch.setattr(
        cmdline,
        "get_output_formats",
        mock.MagicMock(return_value={"homebrew": None}),
    )

    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    with pytest.raises(
        UnsupportedCommandError,
        match=r"The create command for the macOS homebrew format has not been implemented \(yet!\).",
    ):
        do_cmdline_parse("create macOS homebrew".split(), logger, console)


def test_command_explicit_format_help(monkeypatch, capsys, logger, console):
    """``briefcase create macOS app -h`` returns the macOS create app help."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create macOS app -h".split(), logger, console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Help message is for default platform, but app format
    output = capsys.readouterr().out
    assert output.startswith(
        "usage: briefcase create macOS app [-h] [-C KEY=VALUE] [-v] [-V] [--no-input]\n"
        "                                  [--log]\n"
        "\n"
        "Create and populate a macOS app.\n"
    )


def test_command_disable_input(monkeypatch, logger, console):
    """``briefcase create --no-input`` disables console input."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    cmd, options, overrides = do_cmdline_parse(
        "create --no-input".split(), logger, console
    )

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert not cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {}
    assert overrides == {}


def test_command_options(monkeypatch, capsys, logger, console):
    """Commands can provide their own arguments."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    # Invoke a command that is known to have its own custom arguments
    # (In this case, the channel argument for publication)
    cmd, options, overrides = do_cmdline_parse(
        "publish macos app -c s3".split(), logger, console
    )

    assert isinstance(cmd, macOSAppPublishCommand)
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {"channel": "s3"}
    assert overrides == {}


def test_command_overrides(monkeypatch, capsys, logger, console):
    """Configuration overrides can be specified."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    # Invoke a command that is known to have its own custom arguments
    # (In this case, the channel argument for publication)
    cmd, options, overrides = do_cmdline_parse(
        "publish macos app -C version='1.2.3' -C extra=42".split(),
        logger,
        console,
    )

    assert isinstance(cmd, macOSAppPublishCommand)
    assert cmd.input.enabled
    assert cmd.logger.verbosity == LogLevel.INFO
    assert cmd.logger is logger
    assert cmd.input is console
    assert options == {"channel": "s3"}
    assert overrides == {
        "version": "1.2.3",
        "extra": 42,
    }


def test_unknown_command_options(monkeypatch, capsys, logger, console):
    """Commands can provide their own arguments."""
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")

    # Invoke a command but provide an option. that isn't defined
    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("publish macOS xcode -x foobar".split(), logger, console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 2
    # Help message is for default platform and format
    output = capsys.readouterr().err

    assert output.startswith(
        "usage: briefcase publish macOS Xcode [-h] [-C KEY=VALUE] [-v] [-V]\n"
        "                                     [--no-input] [--log] [-c {s3}]\n"
        "briefcase publish macOS Xcode: error: unrecognized arguments: -x"
    )



def test_packaging_formats(default_package_command):
    assert default_package_command.packaging_formats == ["default"]


def test_default_packaging_format(default_package_command):
    assert default_package_command.default_packaging_format == "default"



import importlib
import locale
import os
import platform
import shutil
import sys
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import requests
from cookiecutter.main import cookiecutter

import briefcase.integrations
from briefcase.console import Console, Log
from briefcase.integrations.base import ToolCache

from .test_tool_registry import integrations_modules, tools_for_module


def test_toolcache_typing():
    """Tool typing for ToolCache is correct."""
    # Tools that are intentionally not annotated in ToolCache.
    tools_unannotated = {"cookiecutter"}
    # Tool names to exclude from the dynamic annotation checks; they are manually checked.
    tool_names_skip_dynamic_check = {
        "app_context",  # Tested by the Docker module
        "git",  # An external API, not a Briefcase Tool
        "xcode_cli",  # Tested by the Xcode module
        "ETC_OS_RELEASE",  # A constant, not a tool
    }
    # Tool classes to exclude from dynamic annotation checks.
    tool_klasses_skip_dynamic_checks = {
        "Git",
        "DockerAppContext",
        "NativeAppContext",
        "LinuxDeployQtPlugin",
        "LinuxDeployGtkPlugin",
        "LinuxDeployURLPlugin",
        "LinuxDeployLocalFilePlugin",
    }

    # Ensure all modules containing Tools are exported in ``briefcase.integrations``.
    assert sorted(integrations_modules()) == sorted(briefcase.integrations.__all__)

    # Ensure defined Tool modules/classes are annotated in ToolCache.
    for tool_module_name in briefcase.integrations.__all__:
        if tool_module_name not in tools_unannotated:
            assert tool_module_name in ToolCache.__annotations__
        for tool_name in tools_for_module(tool_module_name):
            if tool_name not in tool_klasses_skip_dynamic_checks:
                assert tool_name in ToolCache.__annotations__.values()

    # Ensure annotated tools use valid Tool names.
    for tool_name, tool_klass_name in ToolCache.__annotations__.items():
        if tool_name not in tool_names_skip_dynamic_check:
            assert tool_name in briefcase.integrations.__all__
            assert tool_klass_name in tools_for_module(tool_name)
            tool_klass = getattr(
                importlib.import_module(f"briefcase.integrations.{tool_name}"),
                tool_klass_name,
            )
            assert tool_name == tool_klass.name

    # Manually check tools that aren't Tool classes or use special annotations.
    app_context_klasses = [
        briefcase.integrations.docker.DockerAppContext.__name__,
        briefcase.integrations.subprocess.Subprocess.__name__,
    ]
    app_context_annotated = ToolCache.__annotations__["app_context"].split(" | ")
    assert sorted(app_context_klasses) == sorted(app_context_annotated)

    assert ToolCache.__annotations__["git"] == "git_"
    assert ToolCache.__annotations__["xcode_cli"] == "XcodeCliTools"
    assert ToolCache.__annotations__["ETC_OS_RELEASE"] == "Path"


def test_third_party_tools_available():
    """Third party tools are available."""
    assert ToolCache.os is os
    assert ToolCache.platform is platform
    assert ToolCache.shutil is shutil
    assert ToolCache.sys is sys

    assert ToolCache.cookiecutter is cookiecutter
    assert ToolCache.requests is requests


def test_always_true(simple_tools, tmp_path):
    """Implicit boolean casts are always True."""
    assert simple_tools or False
    simple_tools["app-1"].app_context = "tool"
    assert simple_tools["app-1"] or False


def test_mapping_protocol(simple_tools):
    """ToolCache is a mapping."""
    simple_tools["app-1"].tool = "tool 1"
    simple_tools["app-2"].tool = "tool 2"

    assert [app for app in simple_tools] == ["app-1", "app-2"]
    assert len(simple_tools) == 2
    assert simple_tools["app-1"].tool == "tool 1"
    assert simple_tools["app-2"].tool == "tool 2"


def test_host_arch_and_os(simple_tools):
    """Arch and OS represent host arch and OS."""
    assert simple_tools.host_arch == platform.machine()
    assert simple_tools.host_os == platform.system()


def test_base_path_is_path(simple_tools):
    """Base path is always a Path."""
    # The BaseCommand tests have much more extensive tests for this path.
    assert isinstance(simple_tools.base_path, Path)
    tools = ToolCache(
        logger=Log(),
        console=Console(),
        base_path="/home/data",
    )
    assert isinstance(tools.base_path, Path)


def test_home_path_default(simple_tools):
    """Home path default is current user's home directory."""
    assert simple_tools.home_path == Path.home()


@pytest.mark.skipif(platform.system() == "Windows", reason="Linux/macOS specific tests")
@pytest.mark.parametrize(
    "home_path, expected_path",
    [
        (None, Path.home()),
        ("/path/to/home", Path("/path/to/home")),
        ("~", Path.home()),
        ("~/dir", Path.home() / "dir"),
    ],
)
def test_nonwindows_home_path(home_path, expected_path, tmp_path):
    """Home path is always expanded or defaulted."""
    tools = ToolCache(
        logger=Log(),
        console=Console(),
        base_path=tmp_path,
        home_path=home_path,
    )
    assert tools.home_path == expected_path


@pytest.mark.skipif(platform.system() != "Windows", reason="Windows specific tests")
@pytest.mark.parametrize(
    "home_path, expected_path",
    [
        (None, Path.home()),
        ("Y:\\path\\to\\home", Path("Y:\\path\\to\\home")),
        ("~", Path.home()),
        ("~/dir", Path.home() / "dir"),
    ],
)
def test_windows_home_path(home_path, expected_path, tmp_path):
    """Home path is always expanded or defaulted."""
    tools = ToolCache(
        logger=Log(),
        console=Console(),
        base_path=tmp_path,
        home_path=home_path,
    )
    assert tools.home_path == expected_path


@pytest.mark.parametrize("maxsize, is_32bit", [(2**32, True), (2**64, False)])
def test_is_32bit_python(maxsize, is_32bit, monkeypatch, tmp_path):
    """Whether Python is 32bits is sensitive to `sys.maxsize`."""
    monkeypatch.setattr(sys, "maxsize", maxsize)

    tools = ToolCache(
        logger=Log(),
        console=Console(),
        base_path=tmp_path,
    )

    assert tools.is_32bit_python is is_32bit


@pytest.mark.parametrize(
    "mock_encoding, expected_encoding",
    [
        ("iso-123", "ISO-123"),
        ("", "ISO-4242"),
        (None, "ISO-4242"),
    ],
)
def test_system_encoding(simple_tools, mock_encoding, expected_encoding, monkeypatch):
    """The expected system encoding is returned."""
    if sys.version_info < (3, 11):
        monkeypatch.setattr(
            locale, "getdefaultlocale", MagicMock(return_value=("aa_BB", mock_encoding))
        )
    else:
        monkeypatch.setattr(
            locale, "getencoding", MagicMock(return_value=mock_encoding)
        )
    monkeypatch.setattr(
        briefcase.integrations.base, "DEFAULT_SYSTEM_ENCODING", "ISO-4242"
    )
    assert simple_tools.system_encoding == expected_encoding



import pytest

from briefcase.config import GlobalConfig
from briefcase.exceptions import BriefcaseConfigError


def test_minimal_GlobalConfig():
    """A simple config can be defined."""
    config = GlobalConfig(
        project_name="My Project",
        version="1.2.3",
        bundle="org.beeware",
        license={"file": "LICENSE"},
    )

    # The basic properties have been set.
    assert config.project_name == "My Project"
    assert config.version == "1.2.3"
    assert config.bundle == "org.beeware"

    assert repr(config) == "<My Project v1.2.3 GlobalConfig>"


def test_extra_attrs():
    """A config can contain attributes in addition to those required."""
    config = GlobalConfig(
        project_name="My Project",
        version="1.2.3",
        bundle="org.beeware",
        url="https://example.com",
        author="Jane Smith",
        author_email="jane@example.com",
        first="value 1",
        second=42,
        license={"file": "LICENSE"},
    )

    # The basic properties have been set.
    assert config.project_name == "My Project"
    assert config.version == "1.2.3"
    assert config.bundle == "org.beeware"
    assert config.url == "https://example.com"
    assert config.author == "Jane Smith"
    assert config.author_email == "jane@example.com"

    # Explicit additional properties have been set
    assert config.first == "value 1"
    assert config.second == 42

    # An attribute that wasn't provided raises an error
    with pytest.raises(AttributeError):
        _ = config.unknown


def test_valid_app_version():
    try:
        GlobalConfig(
            project_name="My Project",
            version="1.2.3",
            bundle="org.beeware",
            license={"file": "LICENSE"},
        )
    except BriefcaseConfigError:
        pytest.fail("1.2.3 should be a valid version number")


def test_invalid_app_version():
    with pytest.raises(
        BriefcaseConfigError, match=r"Version number \(foobar\) is not valid\."
    ):
        GlobalConfig(
            project_name="My Project",
            version="foobar",
            bundle="org.beeware",
            license={"file": "LICENSE"},
        )



from briefcase.commands.base import full_options


def test_no_state():
    """If there's no state, just kwargs are returned."""
    assert full_options(None, {"a": 1, "b": 2}) == {"a": 1, "b": 2}


def test_state():
    """If there's state, it is added to kwargs."""
    assert full_options({"c": 3, "d": 4}, {"a": 1, "b": 2}) == {
        "a": 1,
        "b": 2,
        "c": 3,
        "d": 4,
    }


def test_state_with_overlap():
    """If there's overlap between state and kwargs, state takes precedence."""
    assert full_options({"a": 3, "d": 4}, {"a": 1, "b": 2}) == {"a": 3, "b": 2, "d": 4}



import json
import shlex
from subprocess import CalledProcessError
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError, InvalidDeviceError


@pytest.mark.parametrize(
    "passthrough",
    [
        [],
        [""],
        ['"'],
        ["\\"],
        ["one"],
        ["one two"],
        ["one\ntwo"],
        ["one,two"],
        ["one", "two"],
        ["one", "two", "three"],
    ],
)
def test_start_app_launches_app(adb, capsys, passthrough):
    """Invoking `start_app()` calls `run()` with the appropriate parameters."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(return_value="example normal adb output")

    # Invoke start_app
    adb.start_app(
        "com.example.sample.package", "com.example.sample.activity", passthrough
    )

    # Validate call parameters.
    adb.run.assert_called_once_with(
        "shell",
        "am",
        "start",
        "-n",
        "com.example.sample.package/com.example.sample.activity",
        "-a",
        "android.intent.action.MAIN",
        "-c",
        "android.intent.category.LAUNCHER",
        "--es",
        "org.beeware.ARGV",
        shlex.quote(json.dumps(passthrough)),
    )

    # Validate that the normal output of the command was not printed (since there
    # was no error).
    assert "normal adb output" not in capsys.readouterr()


def test_missing_activity(adb):
    """If the activity doesn't exist, the error is caught."""
    # Use real `adb` output from launching an activity that does not exist.
    # Mock out the run command on an adb instance
    adb.run = MagicMock(
        return_value="""\
Starting: Intent { act=android.intent.action.MAIN cat=[android.intent.category.\
LAUNCHER] cmp=com.example.sample.package/.MainActivity }
Error type 3
Error: Activity class {com.example.sample.package/com.example.sample.package.\
MainActivity} does not exist.
"""
    )

    with pytest.raises(BriefcaseCommandError) as exc_info:
        adb.start_app("com.example.sample.package", "com.example.sample.activity", [])

    assert "Activity class not found" in str(exc_info.value)


def test_invalid_device(adb):
    """If the device doesn't exist, the error is caught."""
    # Use real `adb` output from launching an activity that does not exist.
    # Mock out the run command on an adb instance
    adb.run = MagicMock(side_effect=InvalidDeviceError("device", "exampleDevice"))

    with pytest.raises(InvalidDeviceError):
        adb.start_app("com.example.sample.package", "com.example.sample.activity", [])


def test_unable_to_start(adb):
    """If the adb calls for other reasons, the error is caught."""
    adb.run = MagicMock(side_effect=CalledProcessError(cmd=["adb"], returncode=1))

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to start com.example.sample.package/com.example.sample.activity on exampleDevice",
    ):
        adb.start_app("com.example.sample.package", "com.example.sample.activity", [])



from __future__ import annotations

import json
import os
import re
import shlex
import shutil
import subprocess
import time
from datetime import datetime
from pathlib import Path

from briefcase.config import PEP508_NAME_RE
from briefcase.console import InputDisabled, select_option
from briefcase.exceptions import (
    BriefcaseCommandError,
    IncompatibleToolError,
    InvalidDeviceError,
    MissingToolError,
)
from briefcase.integrations.base import ManagedTool, ToolCache
from briefcase.integrations.java import JDK
from briefcase.integrations.subprocess import SubprocessArgT

DEVICE_NOT_FOUND = re.compile(r"^error: device '[^']*' not found")


class AndroidDeviceNotAuthorized(BriefcaseCommandError):
    def __init__(self, device):
        self.device = device
        super().__init__(
            f"""
The device you have selected ({device}) has not had developer options and
USB debugging enabled. These must be enabled before a device can be used  as a
target for deployment. For details on how to enable Developer Options, visit:

    https://developer.android.com/studio/debug/dev-options#enable

Once you have enabled these options on your device, you will be able to select
this device as a deployment target.

"""
        )


class AndroidSDK(ManagedTool):
    name = "android_sdk"
    full_name = "Android SDK"

    # Latest version for Command-Line Tools download as of May 2024
    # **Be sure the gradle.rst docs stay in sync with version updates here**
    SDK_MANAGER_DOWNLOAD_VER = "11076708"
    SDK_MANAGER_VER = "12.0"

    def __init__(self, tools: ToolCache, root_path: Path):
        super().__init__(tools=tools)
        self.dot_android_path = self.tools.home_path / ".android"
        self.root_path = root_path

        # A wrapper for testing purposes
        self.sleep = time.sleep

    @property
    def cmdline_tools_url(self) -> str:
        """The Android SDK Command-Line Tools URL appropriate for the current machine.

        The SDK largely only supports typical development environments; if a machine is
        using an unsupported architecture, `sdkmanager` will error while installing the
        emulator as a dependency of the build-tools. However, for some of the platforms
        that are unsupported by sdkmanager, users can set up their own SDK install.
        """
        try:
            platform_name = {
                "Darwin": {
                    "arm64": "mac",
                    "x86_64": "mac",
                },
                "Linux": {
                    "x86_64": "linux",
                },
                "Windows": {
                    "AMD64": "win",
                },
            }[self.tools.host_os][self.tools.host_arch]
        except KeyError as e:
            raise IncompatibleToolError(
                tool=self.full_name, env_var="ANDROID_HOME"
            ) from e

        return (
            f"https://dl.google.com/android/repository/"
            f"commandlinetools-{platform_name}-{self.SDK_MANAGER_DOWNLOAD_VER}_latest.zip"
        )

    @property
    def cmdline_tools_path(self) -> Path:
        """Version-specific Command-line tools install root directory."""
        return self.root_path / "cmdline-tools" / self.SDK_MANAGER_VER

    @property
    def sdkmanager_filename(self) -> str:
        return "sdkmanager.bat" if self.tools.host_os == "Windows" else "sdkmanager"

    @property
    def sdkmanager_path(self) -> Path:
        return self.cmdline_tools_path / "bin" / self.sdkmanager_filename

    @property
    def adb_path(self) -> Path:
        adb = "adb.exe" if self.tools.host_os == "Windows" else "adb"
        return self.root_path / "platform-tools" / adb

    @property
    def avdmanager_path(self) -> Path:
        avdmanager = (
            "avdmanager.bat" if self.tools.host_os == "Windows" else "avdmanager"
        )
        return self.cmdline_tools_path / "bin" / avdmanager

    @property
    def emulator_path(self) -> Path:
        emulator = "emulator.exe" if self.tools.host_os == "Windows" else "emulator"
        return self.root_path / "emulator" / emulator

    @property
    def avd_path(self) -> Path:
        return self.dot_android_path / "avd"

    def avd_config_filename(self, avd: str) -> Path:
        return self.avd_path / f"{avd}.avd/config.ini"

    @property
    def env(self) -> dict[str, str]:
        return {
            "ANDROID_HOME": os.fsdecode(self.root_path),
            "ANDROID_SDK_ROOT": os.fsdecode(self.root_path),
            "JAVA_HOME": str(self.tools.java.java_home),
        }

    @property
    def emulator_abi(self) -> str:
        """The ABI to use for the Android emulator."""
        try:
            return {
                "Linux": {
                    "x86_64": "x86_64",
                    "aarch64": "arm64-v8a",
                },
                "Darwin": {
                    "x86_64": "x86_64",
                    "arm64": "arm64-v8a",
                },
                "Windows": {
                    "AMD64": "x86_64",
                },
            }[self.tools.host_os][self.tools.host_arch]
        except KeyError:
            raise BriefcaseCommandError(
                "The Android emulator does not currently support "
                f"{self.tools.host_os} {self.tools.host_arch} hardware."
            )

    @property
    def DEFAULT_DEVICE_TYPE(self) -> str:
        return "pixel"

    @property
    def DEFAULT_DEVICE_SKIN(self) -> str:
        return "pixel_7_pro"

    @property
    def DEFAULT_SYSTEM_IMAGE(self) -> str:
        return f"system-images;android-31;default;{self.emulator_abi}"

    @classmethod
    def sdk_path_from_env(cls, tools: ToolCache) -> tuple[str | None, str | None]:
        """Determine the file path to an Android SDK from the environment.

        Android has historically supported several env vars to set the location of an
        Android SDK for build tools. The currently preferred source is ANDROID_HOME;
        however, ANDROID_SDK_ROOT is also supported as a deprecated setting.

        These values must be same if both set; otherwise, Gradle will error.

        :param tools: ToolCache of available tools
        :returns: Tuple of path to SDK and the env var name that provided that path
        """
        android_home = tools.os.environ.get("ANDROID_HOME")
        android_sdk_root = tools.os.environ.get("ANDROID_SDK_ROOT")

        if android_home:
            if android_sdk_root and android_sdk_root != android_home:
                tools.logger.warning(
                    f"""
*************************************************************************
** WARNING: ANDROID_HOME and ANDROID_SDK_ROOT are inconsistent         **
*************************************************************************

    The ANDROID_HOME and ANDROID_SDK_ROOT environment variables are set
    to different paths:

        ANDROID_HOME:     {android_home}
        ANDROID_SDK_ROOT: {android_sdk_root}

    Briefcase will ignore ANDROID_SDK_ROOT and only use the path
    specified by ANDROID_HOME.

    You should update your environment configuration to either not set
    ANDROID_SDK_ROOT, or set both environment variables to the same
    path.

*************************************************************************
"""
                )
            sdk_root = android_home
            sdk_source = "ANDROID_HOME"
        elif android_sdk_root:
            sdk_root = android_sdk_root
            sdk_source = "ANDROID_SDK_ROOT"
        else:
            sdk_root = None
            sdk_source = None

        return sdk_root, sdk_source

    @classmethod
    def verify_install(
        cls,
        tools: ToolCache,
        install: bool = True,
        **kwargs,
    ) -> AndroidSDK:
        """Verify an Android SDK is available.

        The file paths in ANDROID_HOME and ANDROID_SDK_ROOT environment variables will
        be checked for a valid SDK.

        If those file paths do not contain an SDK, or no file path is provided, an SDK
        is downloaded.

        :param tools: ToolCache of available tools
        :param install: Should the tool be installed if it is not found?
        :returns: A valid Android SDK wrapper. If Android SDK is not available, and was
            not installed, raises MissingToolError.
        """
        # short circuit since already verified and available
        if hasattr(tools, "android_sdk"):
            return tools.android_sdk

        JDK.verify(tools=tools, install=install)

        sdk = None

        # Verify externally-managed Android SDK
        sdk_root_env, sdk_source_env = cls.sdk_path_from_env(tools=tools)
        if sdk_root_env:
            tools.logger.debug("Evaluating ANDROID_HOME...", prefix=cls.full_name)
            tools.logger.debug(f"{sdk_source_env}={sdk_root_env}")
            sdk = AndroidSDK(tools=tools, root_path=Path(sdk_root_env))

            if sdk.exists():
                if sdk_source_env == "ANDROID_SDK_ROOT":
                    tools.logger.warning(
                        """
*************************************************************************
** WARNING: Using Android SDK from ANDROID_SDK_ROOT                    **
*************************************************************************

    Briefcase is using the Android SDK specified by the ANDROID_SDK_ROOT
    environment variable.

    Android has deprecated ANDROID_SDK_ROOT in favor of the
    ANDROID_HOME environment variable.

    Update your environment configuration to set ANDROID_HOME instead of
    ANDROID_SDK_ROOT to ensure future compatibility.

*************************************************************************
"""
                    )
            elif sdk.cmdline_tools_path.parent.exists():
                # a cmdline-tools directory exists but the required version isn't installed.
                # try to install the required version using the 'latest' version.
                if not sdk.install_cmdline_tools():
                    sdk = None
                    tools.logger.warning(
                        f"""
*************************************************************************
** WARNING: Incompatible Command-Line Tools Version                    **
*************************************************************************

    The Android SDK specified by {sdk_source_env} at:

    {sdk_root_env}

    does not contain Command-Line Tools version {cls.SDK_MANAGER_VER}. Briefcase requires
    this version to be installed to use an external Android SDK.

    Use Android Studio's SDK Manager to install Command-Line Tools {cls.SDK_MANAGER_VER}.

    Briefcase will proceed using its own SDK instance.

*************************************************************************
"""
                    )
            else:
                tools.logger.warning(
                    f"""
*************************************************************************
** {f"WARNING: {sdk_source_env} does not point to an Android SDK":67} **
*************************************************************************

    The location pointed to by the {sdk_source_env} environment
    variable:

    {sdk_root_env}

    doesn't appear to contain an Android SDK with the Command-line Tools installed.

    If {sdk_source_env} is an Android SDK, ensure it is the root directory
    of the Android SDK instance such that

    ${sdk_source_env}{os.sep}{sdk.sdkmanager_path.relative_to(sdk.root_path)}

    is a valid filepath.

    Briefcase will proceed using its own SDK instance.

*************************************************************************
"""
                )
                sdk = None

        # Verify Briefcase-managed Android SDK
        if sdk is None:
            sdk_root_path = tools.base_path / "android_sdk"
            sdk = AndroidSDK(tools=tools, root_path=sdk_root_path)

            if not sdk.exists():
                if not install:
                    raise MissingToolError("Android SDK")

                sdk.delete_legacy_sdk_tools()

                if sdk.cmdline_tools_path.parent.exists():
                    tools.logger.info("Upgrading Android SDK...", prefix=cls.name)
                else:
                    tools.logger.info(
                        "The Android SDK was not found; downloading and installing...",
                        prefix=cls.name,
                    )
                    tools.logger.info(
                        "To use an existing Android SDK instance, specify its root "
                        "directory path in the ANDROID_HOME environment variable."
                    )
                    tools.logger.info()
                sdk.install()

        # Licences must be accepted to use the SDK
        sdk.verify_license()

        tools.logger.debug(f"Using Android SDK at {sdk.root_path}")
        tools.android_sdk = sdk
        return sdk

    def exists(self) -> bool:
        """Confirm that the SDK actually exists.

        Look for the sdkmanager; and, if necessary, confirm that it is executable.
        """
        return self.sdkmanager_path.is_file() and (
            self.tools.host_os == "Windows"
            or self.tools.os.access(self.sdkmanager_path, self.tools.os.X_OK)
        )

    @property
    def managed_install(self) -> bool:
        """Is the Android SDK install managed by Briefcase?"""
        # Although the end-user can provide their own SDK, the SDK also
        # provides a built-in upgrade mechanism. Therefore, all Android SDKs
        # are managed installs.
        return True

    def uninstall(self):
        """The Android SDK is upgraded in-place instead of being reinstalled."""

    def install(self):
        """Download and install the Android SDK."""
        cmdline_tools_zip_path = self.tools.file.download(
            url=self.cmdline_tools_url,
            download_path=self.tools.base_path,
            role="Android SDK Command-Line Tools",
        )

        # The cmdline-tools package *must* be installed as:
        #     <sdk_path>/cmdline-tools/<cmdline-tools version>
        #
        # However, the zip file unpacks a top-level folder named `cmdline-tools`.
        # So, the unpacking process is:
        #
        #  1. Make a <sdk_path>/cmdline-tools folder
        #  2. Unpack the zip file into that folder, creating <sdk_path>/cmdline-tools/cmdline-tools
        #  3. Move <sdk_path>/cmdline-tools/cmdline-tools to <sdk_path>/cmdline-tools/<cmdline-tools version>

        with self.tools.input.wait_bar(
            f"Installing Android SDK Command-Line Tools {self.SDK_MANAGER_VER}..."
        ):
            self.cmdline_tools_path.parent.mkdir(parents=True, exist_ok=True)
            try:
                self.tools.file.unpack_archive(
                    cmdline_tools_zip_path, extract_dir=self.cmdline_tools_path.parent
                )
            except (shutil.ReadError, EOFError) as e:
                raise BriefcaseCommandError(
                    f"""\
Unable to unpack Android SDK Command-Line Tools ZIP file. The download may have been interrupted
or corrupted.

Delete {cmdline_tools_zip_path} and run briefcase again.
"""
                ) from e

            # If there's an existing version of the cmdline tools, delete them.
            if self.cmdline_tools_path.exists():
                self.tools.shutil.rmtree(self.cmdline_tools_path)

            # Rename the top level zip content to the final name
            (self.cmdline_tools_path.parent / "cmdline-tools").rename(
                self.cmdline_tools_path
            )

            # Zip file no longer needed once unpacked.
            cmdline_tools_zip_path.unlink()

            # Python zip unpacking ignores permission metadata.
            # On non-Windows, we manually fix permissions.
            if (  # pragma: no branch
                self.tools.host_os != "Windows"
            ):  # pragma: no-cover-if-is-windows
                for binpath in (self.cmdline_tools_path / "bin").glob("*"):
                    if not self.tools.os.access(binpath, self.tools.os.X_OK):
                        binpath.chmod(0o755)

        with self.tools.input.wait_bar("Removing older Android SDK packages..."):
            self.cleanup_old_installs()

    def upgrade(self):
        """Upgrade the Android SDK."""
        try:
            # Using subprocess.run() with no I/O redirection so the user sees
            # the full output and can send input.
            self.tools.subprocess.run(
                [self.sdkmanager_path, "--update"],
                env=self.env,
                check=True,
                stream_output=False,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"""\
Error while updating the Android SDK manager. Please run this command and examine
its output for errors.

    $ {self.sdkmanager_path} --update
"""
            ) from e

    def install_cmdline_tools(self) -> bool:
        """Attempt to use 'latest' cmdline-tools to install the currently required
        version of the Command-Line Tools.

        The Briefcase-managed SDK should always have the required version of cmdline-
        tools installed; however, user-provided SDKs may not have it.

        :returns: True if successfully installed; False otherwise
        """
        self.tools.logger.info(
            f"Installing Android Command-Line Tools {self.SDK_MANAGER_VER}...",
            prefix=self.full_name,
        )
        self.tools.logger.info(f"Using Android SDK at {self.root_path}")
        latest_sdkmanager_path = (
            self.root_path
            / "cmdline-tools"
            / "latest"
            / "bin"
            / self.sdkmanager_filename
        )
        try:
            self.tools.subprocess.run(
                [
                    latest_sdkmanager_path,
                    f"cmdline-tools;{self.SDK_MANAGER_VER}",
                ],
                check=True,
                stream_output=False,
            )
        except (OSError, subprocess.CalledProcessError) as e:
            self.tools.logger.debug(str(e))
            self.tools.logger.warning(
                f"Failed to install cmdline-tools;{self.SDK_MANAGER_VER}"
            )
            return False
        return True

    def delete_legacy_sdk_tools(self):
        """Delete any legacy Android SDK tools that are installed.

        If no versions of the Command-Line Tools are installed but the 'tools' directory
        exists, the legacy SDK Tools are probably installed. Since they have been
        deprecated by more recent releases of SDK Manager, delete them and perform a
        fresh install.

        The Android SDK Tools were deprecated in Sept 2017.
        """
        if (
            not self.cmdline_tools_path.parent.exists()
            and (self.root_path / "tools").exists()
        ):
            self.tools.logger.warning(
                f"""
*************************************************************************
** WARNING: Upgrading Android SDK tools                                **
*************************************************************************

    Briefcase needs to replace the older Android SDK Tools with the
    newer Android SDK Command-Line Tools. This will involve some large
    downloads, as well as re-accepting the licenses for the Android
    SDKs.

    Any emulators created with the older Android SDK Tools will not be
    compatible with the new tools. You will need to create new
    emulators. Old emulators can be removed by deleting the files
    in {self.avd_path} matching the emulator name.

*************************************************************************
"""
            )
            self.tools.shutil.rmtree(self.root_path)

    def cleanup_old_installs(self):
        """Remove old versions of Android SDK packages and version markers.

        When the Android SDK is upgraded, old versions of packages should be removed to
        keep the SDK tidy. This is namely the Command-line Tools that are used to manage
        the SDK and AVDs. Additionally, previous version of Briefcase created a version
        marker file that needs to be deleted.
        """
        if (ver_file := self.cmdline_tools_path.parent / "8092744").is_file():
            self.tools.os.unlink(ver_file)
        if (latest := self.cmdline_tools_path.parent / "latest").is_dir():
            self.tools.shutil.rmtree(latest)

    def list_packages(self):
        """In debug output, list the packages currently managed by the SDK."""
        try:
            # check_output always writes its output to debug
            self.tools.subprocess.check_output(
                [self.sdkmanager_path, "--list_installed"],
                env=self.env,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                "Unable to invoke the Android SDK manager"
            ) from e

    def adb(self, device: str) -> ADB:
        """Obtain an ADB instance for managing a specific device.

        :param device: The device ID to manage.
        """
        return ADB(tools=self.tools, device=device)

    def verify_license(self):
        """Verify that all necessary licenses have been accepted.

        If they haven't, prompt the user to do so.

        Raises an error if licenses are not.
        """
        license_path = self.root_path / "licenses/android-sdk-license"
        if license_path.exists():
            return

        self.tools.logger.info(
            """
The Android tools provided by Google have license terms that you must accept
before you may use those tools.
"""
        )
        try:
            # Using subprocess.run() with no I/O redirection so the user sees
            # the full output and can send input.
            self.tools.subprocess.run(
                [self.sdkmanager_path, "--licenses"],
                env=self.env,
                check=True,
                stream_output=False,
            )
        except (subprocess.CalledProcessError, OSError) as e:
            raise BriefcaseCommandError(
                f"""\
Error while reviewing Android SDK licenses. Please run this command and examine
its output for errors.

    $ {self.sdkmanager_path} --licenses
"""
            ) from e

        if not license_path.exists():
            raise BriefcaseCommandError(
                """\
You did not accept the Android SDK licenses. Please re-run the briefcase command
and accept the Android SDK license when prompted. You may need an Internet
connection.
"""
            )

    def verify_emulator(self):
        """Verify that Android emulator has been installed, and is in a runnable state.

        Raises an error if the emulator can't be installed.
        """
        # Ensure the `platforms` folder exists.
        # See the discussion on #766 for details; as of June 2022, if this folder
        # doesn't exist, the emulator won't start, raising the error:
        #
        #    PANIC: Cannot find AVD system path. Please define ANDROID_SDK_ROOT
        #
        # Creating an empty platforms folder is enough to overcome this. This folder
        # will be created automatically when you build a project; but if you have a
        # clean Android SDK install that hasn't been used to build a project, it
        # might be missing.
        (self.root_path / "platforms").mkdir(exist_ok=True)

        if self.emulator_path.exists():
            self.tools.logger.debug("Android emulator is already installed.")
            return

        self.tools.logger.info("Downloading the Android emulator...")
        try:
            self.tools.subprocess.run(
                [self.sdkmanager_path, "platform-tools", "emulator"],
                env=self.env,
                check=True,
                stream_output=False,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                "Error while installing Android emulator."
            ) from e

    def verify_avd(self, avd: str):
        """Verify that the AVD has the necessary system components to launch.

        This includes:
            * AVD system image
            * Emulator skin

        :param avd: The AVD name to verify.
        """
        # Read the AVD configuration to retrieve the system image.
        # This is stored in the AVD configuration file with the key:
        #   image.sysdir.1=system-images/android-31/default/arm64-v8a/
        avd_config = self.avd_config(avd)

        try:
            system_image_path = Path(avd_config["image.sysdir.1"])

            # Convert the path into a system image name, and verify it.
            self.verify_system_image(";".join(system_image_path.parts))
        except KeyError:
            self.tools.logger.warning(
                f"""
*************************************************************************
** WARNING: Unable to determine AVD system image                       **
*************************************************************************

    Briefcase was unable to determine the system image of the Android
    emulator AVD {avd!r} from it's configuration file.

    Briefcase will proceed assuming the emulator is correctly
    configured. If you experience any problems running the emulator,
    this may be the cause of the problem.

*************************************************************************
"""
            )

        try:
            skin = avd_config["skin.name"]
            skin_path = Path(avd_config["skin.path"])
            if skin_path == Path("_no_skin"):
                self.tools.logger.debug("Emulator does not use a skin.")
            elif skin_path != Path("skins") / skin:
                self.tools.logger.warning(
                    f"""
*************************************************************************
** WARNING: Unrecognized device skin                                   **
*************************************************************************

    Briefcase does not recognize the skin {skin!r} used by the
    Android emulator AVD {avd!r}.

    Briefcase will proceed assuming the emulator is correctly
    configured. If you experience any problems running the emulator,
    this may be the cause of the problem.

*************************************************************************
"""
                )
            else:
                # Convert the path into a system image name, and verify it.
                self.verify_emulator_skin(skin)

        except KeyError:
            self.tools.logger.debug(f"Device {avd!r} doesn't define a skin.")

    def verify_system_image(self, system_image: str):
        """Verify that the required system image is installed.

        :param system_image: The SDKManager identifier for the system image (e.g.,
            ``"system-images;android-31;default;x86_64"``)
        """
        # Look for the directory named as a system image.
        # If it exists, we already have the system image.
        system_image_parts = system_image.split(";")

        if len(system_image_parts) < 4 or system_image_parts[0] != "system-images":
            raise BriefcaseCommandError(
                f"{system_image!r} is not a valid system image name."
            )

        if system_image_parts[-1] != self.emulator_abi:
            self.tools.logger.warning(
                f"""
*************************************************************************
** WARNING: Unexpected emulator ABI                                    **
*************************************************************************

    The system image {system_image!r}
    does not match the architecture of this computer ({self.emulator_abi}).

    Briefcase will proceed assuming the emulator is correctly
    configured. If you experience any problems running the emulator,
    this may be the cause of the problem.

*************************************************************************
"""
            )

        # Convert the system image into a path where that system image
        # would be expected, and see if the location exists.
        system_image_path = self.root_path
        for part in system_image_parts:
            system_image_path = system_image_path / part

        if system_image_path.exists():
            # Found the system image.
            return

        # System image not found; download it.
        self.tools.logger.info(
            f"Downloading the {system_image!r} Android system image...",
            prefix=self.name,
        )
        try:
            self.tools.subprocess.run(
                [self.sdkmanager_path, system_image],
                env=self.env,
                check=True,
                stream_output=False,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Error while installing the {system_image!r} Android system image."
            ) from e

    def verify_emulator_skin(self, skin: str):
        """Verify that an emulator skin is available.

        A human-readable list of available skins can be found here:

            https://android.googlesource.com/platform/tools/adt/idea/+/refs/heads/mirror-goog-studio-main/artwork/resources/device-art-resources/

        :param skin: The name of the skin to obtain
        """
        # Check for a device skin. If it doesn't exist, download it.
        skin_path = self.root_path / "skins" / skin
        if skin_path.exists():
            self.tools.logger.debug(f"Device skin {skin!r} already exists.")
            return

        self.tools.logger.info(f"Obtaining {skin} device skin...", prefix=self.name)

        skin_url = (
            "https://android.googlesource.com/platform/tools/adt/idea/"
            "+archive/refs/heads/mirror-goog-studio-main/"
            f"artwork/resources/device-art-resources/{skin}.tar.gz"
        )

        skin_tgz_path = self.tools.file.download(
            url=skin_url,
            download_path=self.root_path,
            role=f"{skin} device skin",
        )

        # Unpack skin archive
        with self.tools.input.wait_bar("Installing device skin..."):
            try:
                self.tools.file.unpack_archive(
                    skin_tgz_path,
                    extract_dir=skin_path,
                )
            except (shutil.ReadError, EOFError) as e:
                raise BriefcaseCommandError(
                    f"Unable to unpack {skin} device skin."
                ) from e

            # Delete the downloaded file.
            skin_tgz_path.unlink()

    def emulators(self) -> list[str]:
        """Find the list of emulators that are available."""
        try:
            emulators = self.tools.subprocess.check_output(
                [self.emulator_path, "-list-avds"]
            ).strip()
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError("Unable to obtain Android emulator list") from e
        else:
            return [
                emu
                for emu in emulators.split("\n")
                # ignore any logging output included in output list
                if emu and not emu.startswith(("INFO    |", "WARNING |", "ERROR   |"))
            ]

    def devices(self) -> dict[str, dict[str, str | bool]]:
        """Find the devices that are attached and available to ADB."""
        try:
            output = self.tools.subprocess.check_output(
                [self.adb_path, "devices", "-l"]
            ).strip()
            # Process the output of `adb devices -l`.
            # The first line is header information.
            # Each subsequent line is a single device descriptor.
            devices = {}
            header_found = False
            for line in output.split("\n"):
                if line == "List of devices attached":
                    header_found = True
                elif header_found and line:
                    parts = re.sub(r"\s+", " ", line).split(" ")

                    details = {}
                    for part in parts[2:]:
                        try:
                            key, value = part.split(":")
                            details[key] = value
                        except ValueError:
                            # Ignore any entry that isn't in "key:value" format.
                            pass

                    if parts[1] == "device":
                        try:
                            name = details["model"].replace("_", " ")
                        except KeyError:
                            name = "Unknown device (no model name)"
                        authorized = True
                    elif parts[1] == "offline":
                        name = "Unknown device (offline)"
                        authorized = False
                    else:
                        name = f"Device not available for development ({' '.join(parts[1:])})"
                        authorized = False

                    devices[parts[0]] = {
                        "name": name,
                        "authorized": authorized,
                    }

            return devices
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError("Unable to obtain Android device list") from e

    def select_target_device(
        self,
        device_or_avd: str | None,
    ) -> tuple[str | None, str | None, str | None]:
        """Select a device to be the target for actions.

        Interrogates the system to get the list of available devices.

        If the user has specified a device at the command line, that device will
        be validated, and then automatically selected.

        :param device_or_avd: The device or AVD to target. Can be a physical
            device id (a hex string), an emulator id (``emulator-5554``), or an
            emulator AVD name (``@robotfriend``), or a JSON payload describing
            the properties of an emulator that will be created (e.g.,
            ``'{"avd":"beePhone","device_type":"pixel","skin":"pixel_3a","system_image":"system-images;android-31;default;arm64-v8a"}'``)
            If ``None``, the user will be asked to select a device from the list
            available.
        :returns: A tuple containing ``(device, name, avd)``. ``avd`` will only
            be provided if an emulator with that AVD is not currently running.
            If ``device`` is None, a new emulator should be created.
        """
        # If the device_or_avd starts with "{", it's a definition for a new
        # emulator to be created.
        if device_or_avd and device_or_avd.startswith("{"):
            try:
                emulator_config = json.loads(device_or_avd)
                emulators = set(self.emulators())

                # If an emulator with this AVD already exists, use it
                avd = emulator_config["avd"]
                if avd not in emulators:
                    self._create_emulator(**emulator_config)

                return None, f"@{avd} (emulator)", avd
            except json.JSONDecodeError as e:
                raise BriefcaseCommandError(
                    f"Unable to create emulator with definition {device_or_avd!r}"
                ) from e
            except KeyError:
                raise BriefcaseCommandError("No AVD provided for new device.")
            except TypeError as e:
                property = str(e).split(" ")[-1]
                raise BriefcaseCommandError(f"Unknown device property {property}.")

        # Get the list of attached devices (includes running emulators)
        running_devices = self.devices()

        # Choices is an ordered list of options that can be shown to the user.
        # Each device should appear only once, and be keyed by AVD only if
        # a device ID isn't available.
        choices = []
        # Device choices is the full lookup list. Devices can be looked up
        # by any valid key - ID *or* AVD.
        device_choices = {}

        # Iterate over all the running devices.
        # If the device is a virtual device, use ADB to get the emulator AVD name.
        # If it is a physical device, use the device name.
        # Keep a log of all running AVDs
        running_avds = {}
        for d, details in sorted(running_devices.items(), key=lambda d: d[1]["name"]):
            name = details["name"]
            avd = self.adb(d).avd_name()
            if avd:
                # It's a running emulator
                running_avds[avd] = d
                full_name = f"@{avd} (running emulator)"
                choices.append((d, full_name))

                # Save the AVD as a device detail.
                details["avd"] = avd

                # Device can be looked up by device ID or AVD
                device_choices[d] = full_name
                device_choices[f"@{avd}"] = full_name
            else:
                # It's a physical device (might be disabled)
                full_name = f"{name} ({d})"
                choices.append((d, full_name))
                device_choices[d] = full_name

        # Add any non-running emulator AVDs to the list of candidate devices
        for avd in self.emulators():
            if avd not in running_avds:
                name = f"@{avd} (emulator)"
                choices.append((f"@{avd}", name))
                device_choices[f"@{avd}"] = name

        # If a device or AVD has been provided, check it against the available
        # device list.
        if device_or_avd:
            try:
                name = device_choices[device_or_avd]

                if device_or_avd.startswith("@"):
                    # specifier is an AVD
                    avd = device_or_avd[1:]
                    try:
                        device = running_avds[avd]
                    except KeyError:
                        # device_or_avd isn't in the list of running avds;
                        # it must be a non-running emulator.
                        return None, name, avd
                else:
                    # Specifier is a direct device ID
                    device = device_or_avd

                details = running_devices[device]
                avd = details.get("avd")
                if details["authorized"]:
                    # An authorized, running device (emulator or physical)
                    return device, name, avd
                else:
                    # An unauthorized physical device
                    raise AndroidDeviceNotAuthorized(device)

            except KeyError as e:
                # Provided device_or_id isn't a valid device identifier.
                id_type = (
                    "emulator AVD" if device_or_avd.startswith("@") else "device ID"
                )
                raise InvalidDeviceError(id_type, device_or_avd) from e
        # We weren't given a device/AVD; we have to select from the list.
        # If we're selecting from a list, there's always one last choice
        choices.append((None, "Create a new Android emulator"))

        # Show the choices to the user.
        self.tools.input.prompt()
        self.tools.input.prompt("Select device:")
        self.tools.input.prompt()
        try:
            choice = select_option(choices, input=self.tools.input)
        except InputDisabled as e:
            # If input is disabled, and there's only one actual simulator,
            # select it. If there are no simulators, select "Create simulator"
            if len(choices) <= 2:
                choice = choices[0][0]
            else:
                raise BriefcaseCommandError(
                    """\
Input has been disabled; can't select a device to target.
Use the -d/--device option to explicitly specify the device to use.
"""
                ) from e

        # Process the user's choice
        if choice is None:
            # Create a new emulator. No device ID or AVD.
            device = None
            avd = None
            name = None
        elif choice.startswith("@"):
            # A non-running emulator. We have an AVD, but no device ID.
            device = None
            name = device_choices[choice]
            avd = choice[1:]
        else:
            # Either a running emulator, or a physical device. Regardless,
            # we need to check if the device is developer enabled.
            # Functionally, we know the device *must* be in the list of
            # choices; which means it's also in the list of running devices
            # and the list of device choices, so any KeyError on those lookups
            # indicates a deeper problem.
            details = running_devices[choice]
            if not details["authorized"]:
                # An unauthorized physical device
                raise AndroidDeviceNotAuthorized(choice)

            # Return the device ID and name.
            device = choice
            name = device_choices[choice]
            avd = details.get("avd")

        if avd:
            self.tools.logger.info(
                f"""
In future, you can specify this device by running:

    $ briefcase run android -d "@{avd}"
"""
            )
        elif device:
            self.tools.logger.info(
                f"""
In future, you can specify this device by running:

    $ briefcase run android -d {device}
"""
            )

        return device, name, avd

    def create_emulator(self) -> str:
        """Create a new Android emulator.

        :returns: The AVD of the newly created emulator.
        """
        # Get the list of existing emulators
        emulators = set(self.emulators())

        default_avd = "beePhone"
        i = 1
        # Make sure the default name is unique
        while default_avd in emulators:
            i += 1
            default_avd = f"beePhone{i}"

        # Prompt for a device avd until a valid one is provided.
        self.tools.logger.info(
            f"""
You need to select a name for your new emulator. This is an identifier that
can be used to start the emulator in future. It should follow the same naming
conventions as a Python package (i.e., it may only contain letters, numbers,
hyphens and underscores). If you don't provide a name, Briefcase will use the
a default name '{default_avd}'.

"""
        )
        avd_is_invalid = True
        while avd_is_invalid:
            avd = self.tools.input(f"Emulator name [{default_avd}]: ")
            # If the user doesn't provide a name, use the default.
            if avd == "":
                avd = default_avd

            if not PEP508_NAME_RE.match(avd):
                self.tools.logger.info(
                    f"""
'{avd}' is not a valid emulator name. An emulator name may only contain
letters, numbers, hyphens and underscores.

"""
                )
            elif avd in emulators:
                self.tools.logger.info(
                    f"""
An emulator named '{avd}' already exists.

"""
                )
            else:
                avd_is_invalid = False

        # TODO: Provide a list of options for device types with matching skins
        device_type = self.DEFAULT_DEVICE_TYPE
        skin = self.DEFAULT_DEVICE_SKIN

        # TODO: Provide a list of options for system images.
        system_image = self.DEFAULT_SYSTEM_IMAGE

        self._create_emulator(
            avd=avd,
            device_type=device_type,
            skin=skin,
            system_image=system_image,
        )

        self.tools.logger.info(
            f"""
Android emulator '{avd}' created.

In future, you can specify this device by running:

    $ briefcase run android -d @{avd}
"""
        )

        return avd

    def _create_emulator(
        self,
        avd: str,
        device_type: str | None = None,
        skin: str | None = None,
        system_image: str | None = None,
    ):
        """Internal method that does the actual work of creating the emulator.

        AVD is the only required argument; all other arguments will assume reasonable
        defaults.

        :param avd: The AVD for the new emulator
        :param device_type: The device type for the new emulator (e.g., "pixel")
        :param skin: The skin for the new emulator to use (e.g., "pixel_3a")
        :param system_image: The system image to use on the new emulator. (e.g.,
            "system-images;android-31;default;arm64-v8a")
        """
        if device_type is None:
            device_type = self.DEFAULT_DEVICE_TYPE
        if skin is None:
            skin = self.DEFAULT_DEVICE_SKIN
        if system_image is None:
            system_image = self.DEFAULT_SYSTEM_IMAGE

        # Ensure the required skin is available.
        self.verify_emulator_skin(skin)

        # Ensure the required system image is available.
        self.verify_system_image(system_image)

        with self.tools.input.wait_bar(f"Creating Android emulator {avd}..."):
            try:
                self.tools.subprocess.check_output(
                    [
                        self.avdmanager_path,
                        "--verbose",
                        "create",
                        "avd",
                        "--name",
                        avd,
                        "--abi",
                        self.emulator_abi,
                        "--package",
                        system_image,
                        "--device",
                        device_type,
                    ],
                    # Ensure XDG_CONFIG_HOME is not set so avdmanager uses the default
                    # location (i.e. ~/.android) because the emulator does not respect
                    # XDG_CONFIG_HOME and will not be able to find the AVD to run it.
                    env={
                        **self.env,
                        **{"XDG_CONFIG_HOME": None},
                    },
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError("Unable to create Android emulator") from e

        with self.tools.input.wait_bar("Adding extra device configuration..."):
            self.update_emulator_config(
                avd,
                {
                    "avd.id": avd,
                    "avd.name": avd,
                    "disk.dataPartition.size": "4096M",
                    "hw.keyboard": "yes",
                    "skin.dynamic": "yes",
                    "skin.name": skin,
                    "skin.path": f"skins/{skin}",
                    "showDeviceFrame": "yes",
                },
            )

    def avd_config(self, avd: str) -> dict[str, str]:
        """Obtain the AVD configuration as key-value pairs.

        :params avd: The AVD whose config will be retrieved
        """
        # Parse the existing config into key-value pairs
        avd_config = {}
        try:
            with self.avd_config_filename(avd).open("r", encoding="utf-8") as f:
                for line in f:
                    try:
                        key, value = line.rstrip().split("=", 1)
                        avd_config[key.strip()] = value.strip()
                    except ValueError:
                        pass
        except OSError as e:
            raise BriefcaseCommandError(
                f"Unable to read configuration of AVD @{avd}"
            ) from e

        return avd_config

    def update_emulator_config(self, avd: str, updates: dict[str, str]):
        """Update the AVD configuration with specific values.

        :params avd: The AVD whose config will be updated
        :params updates: A dictionary containing the new key-value to add to the device
            configuration.
        """
        avd_config = self.avd_config(avd)

        # Augment the config with the new key-values pairs
        avd_config.update(updates)

        # Write the update configuration.
        with self.avd_config_filename(avd).open("w", encoding="utf-8") as f:
            for key, value in avd_config.items():
                f.write(f"{key}={value}\n")

    def start_emulator(
        self,
        avd: str,
        extra_args: list[str] | None = None,
    ) -> tuple[str, str]:
        """Start an existing Android emulator.

        Returns when the emulator is booted and ready to accept apps.

        :param avd: The AVD of the device.
        :param extra_args: Additional command line arguments to pass when starting the
            emulator.
        """
        if avd not in set(self.emulators()):
            raise InvalidDeviceError("emulator AVD", avd)

        if extra_args is None:
            extra_args = []

        # Start the emulator
        emulator_popen = self.tools.subprocess.Popen(
            [self.emulator_path, f"@{avd}", "-dns-server", "8.8.8.8"] + extra_args,
            env=self.env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
            start_new_session=True,
        )

        # Start capturing the emulator's output
        # On Windows, the emulator can block until stdout is read and the emulator will
        # not actually run until the user sends CTRL+C to Briefcase (#1573). This
        # avoids that scenario while also ensuring emulator output is always available
        # to print in the console if other issues occur.
        emulator_streamer = self.tools.subprocess.stream_output_non_blocking(
            label="Android emulator",
            popen_process=emulator_popen,
            capture_output=True,
        )

        # wrap AVD name in quotes since '@' is a special char in PowerShell
        emulator_command = " ".join(
            f'"{arg}"' if arg.startswith("@") else arg
            for arg in map(str, emulator_popen.args)
        )

        general_error_msg = f"""
Review the emulator output above for:
 - Troubleshooting or resolution steps such as enabling hardware acceleration
 - Other errors or warnings that may be suggesting the cause of the startup failure

Ensure your Android SDK is up-to-date by running:

    $ briefcase upgrade {AndroidSDK.name}

To review Google's general troubleshooting steps for the emulator, visit:

    https://developer.android.com/studio/run/emulator-troubleshooting

You can also start the emulator manually by running:

    $ {emulator_command}
"""

        failed_startup_error_msg = f"{{prologue}}\n{general_error_msg}"

        # The boot process happens in 2 phases.
        # First, the emulator appears in the device list. However, it's not ready until
        # the boot process has finished. To determine the boot status, we need the
        # device ID, and an ADB connection.

        # Phase 1: Wait for the device to appear so we can get an ADB instance for it.
        try:
            with self.tools.input.wait_bar("Starting emulator...") as keep_alive:
                adb = None
                known_devices = set()
                while adb is None:
                    if emulator_popen.poll() is not None:
                        raise BriefcaseCommandError(
                            failed_startup_error_msg.format(
                                prologue="Android emulator was unable to start!"
                            )
                        )

                    for device, details in sorted(self.devices().items()):
                        # Only process authorized devices that we haven't seen.
                        if details["authorized"] and device not in known_devices:
                            adb = self.adb(device)
                            device_avd = adb.avd_name()

                            if device_avd == avd:
                                # Found an active device that matches
                                # the AVD we are starting.
                                full_name = f"@{avd} (running emulator)"
                                break
                            else:
                                # Not the one. Zathras knows.
                                adb = None
                                known_devices.add(device)

                    # If we haven't found a device, try again in 2 seconds...
                    if adb is None:
                        keep_alive.update()
                        self.sleep(2)

            # Phase 2: Wait for the boot process to complete
            if not adb.has_booted():
                with self.tools.input.wait_bar("Booting emulator...") as keep_alive:
                    while not adb.has_booted():
                        if emulator_popen.poll() is not None:
                            raise BriefcaseCommandError(
                                failed_startup_error_msg.format(
                                    prologue="Android emulator was unable to boot!"
                                )
                            )

                        # Try again in 2 seconds...
                        keep_alive.update()
                        self.sleep(2)
        except BaseException as e:
            self.tools.logger.warning(
                "Emulator output log for startup failure",
                prefix=self.name,
            )
            self.tools.logger.info(emulator_streamer.captured_output)

            # Provide troubleshooting steps if user gives up on the emulator starting
            if isinstance(e, KeyboardInterrupt):
                self.tools.logger.warning(
                    "Is the Android emulator not starting up properly?",
                    prefix=self.name,
                )
                self.tools.logger.info(
                    """
If the emulator opened after pressing CTRL+C, then leave the emulator open and
run Briefcase again. The running emulator can then be selected from the list.
"""
                )
                self.tools.logger.info(general_error_msg)

            raise
        finally:
            emulator_streamer.request_stop()

        # Return the device ID and full name.
        return device, full_name


class ADB:
    def __init__(self, tools: ToolCache, device: str):
        """An API integration for the Android Debug Bridge (ADB).

        :param tools: ToolCache of available tools
        :param device: The ID of the device to target (in a format usable by `adb -s`)
        """
        self.tools = tools
        self.device = device

    def avd_name(self) -> str | None:
        """Get the AVD name for the device.

        :returns: The AVD name for the device; or ``None`` if the device isn't
            an emulator
        """
        try:
            output = self.run("emu", "avd", "name")
            return output.split("\n")[0]
        except subprocess.CalledProcessError as e:
            # Status code 1 is a normal "it's not an emulator" error response
            if e.returncode == 1:
                return None
            else:
                raise BriefcaseCommandError(
                    f"Unable to interrogate AVD name of device {self.device}"
                ) from e

    def has_booted(self) -> bool:
        """Determine if the device has completed booting.

        :returns: True if it has booted; False otherwise.
        """
        try:
            # When the sys.boot_completed property of the device
            # returns '1', the boot is complete. Any other response indicates
            # booting is underway.
            output = self.run("shell", "getprop", "sys.boot_completed")
            return output.strip() == "1"
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to determine if emulator {self.device} has booted."
            ) from e

    def run(self, *arguments: SubprocessArgT, quiet: bool = False) -> str:
        """Run a command on a device using Android debug bridge, `adb`. The device name
        is mandatory to ensure clarity in the case of multiple attached devices.

        :param arguments: List of strings to pass to `adb` as arguments.
        :param quiet: Should the invocation of this command be silent, and
            *not* appear in the logs? This should almost always be False;
            however, for some calls (most notably, calls that are called
            frequently to evaluate the status of another process), logging can
            be turned off so that log output isn't corrupted by thousands of
            polling calls.
        :returns: `adb` output on success; raises an exception on failure.
        """
        # The ADB integration operates on the basis of running commands before
        # checking that they are valid, then parsing output to notice errors.
        # This keeps performance good in the success case.
        try:
            output = self.tools.subprocess.check_output(
                [self.tools.android_sdk.adb_path, "-s", self.device] + list(arguments),
                quiet=quiet,
            )
            # add returns status code 0 in the case of failure. The only tangible evidence
            # of failure is the message "Failure [INSTALL_FAILED_OLDER_SDK]" in the,
            # console output; so if that message exists in the output, raise an exception.
            if "Failure [INSTALL_FAILED_OLDER_SDK]" in output:
                raise BriefcaseCommandError(
                    "Your device doesn't meet the minimum SDK requirements of this app."
                )
            return output
        except subprocess.CalledProcessError as e:
            if any(DEVICE_NOT_FOUND.match(line) for line in e.output.split("\n")):
                raise InvalidDeviceError("device id", self.device) from e
            raise

    def install_apk(self, apk_path: str | Path):
        """Install an APK file on an Android device.

        :param apk_path: The path of the Android APK file to install.
        :returns: `None` on success; raises an exception on failure.
        """
        try:
            self.run("install", "-r", apk_path)
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to install APK {apk_path} on {self.device}"
            ) from e

    def force_stop_app(self, package: str):
        """Force-stop an app, specified as a package name.

        :param package: The name of the Android package, e.g., com.username.myapp.
        :returns: `None` on success; raises an exception on failure.
        """
        # In my testing, `force-stop` exits with status code 0 (success) so long
        # as you pass a package name, even if the package does not exist, or the
        # package is not running.
        try:
            self.run("shell", "am", "force-stop", package)
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to force stop app {package} on {self.device}"
            ) from e

    def start_app(self, package: str, activity: str, passthrough: list[str]):
        """Start an app, specified as a package name & activity name.

        If you have an APK file, and you are not sure of the package or activity
        name, you can find it using `aapt dump badging filename.apk` and looking
        for "package" and "launchable-activity" in the output.

        :param package: The name of the Android package, e.g., com.username.myapp.
        :param activity: The activity of the APK to start.
        :param passthrough: Arguments to pass to the app.
        :returns: `None` on success; raises an exception on failure.
        """
        try:
            # `am start` also accepts string array extras, but we pass the arguments as a
            # single JSON string, because JSON deals with edge cases like whitespace and
            # escaping in a reliable and well-documented way.
            output = self.run(
                "shell",
                "am",
                "start",
                "-n",
                f"{package}/{activity}",
                "-a",
                "android.intent.action.MAIN",
                "-c",
                "android.intent.category.LAUNCHER",
                "--es",
                "org.beeware.ARGV",
                shlex.quote(json.dumps(passthrough)),  # Protect from Android's shell
            )

            # `adb shell am start` always exits with status zero. We look for error
            # messages in the output.
            if any(
                line.startswith("Error: Activity class ")
                and line.endswith("does not exist.")
                for line in output.split("\n")
            ):
                raise BriefcaseCommandError(
                    f"""\
Activity class not found while starting app.

`adb` output:

    {output}
"""
                )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to start {package}/{activity} on {self.device}"
            ) from e

    def logcat(self, pid: str) -> subprocess.Popen:
        """Start following the adb log for the device.

        :param pid: The PID whose logs you want to display.
        :returns: A Popen object for the logcat call
        """
        # As best as we can make out, adb logcat returns UTF-8 output.
        # See #1425 for details.
        return self.tools.subprocess.Popen(
            [
                self.tools.android_sdk.adb_path,
                "-s",
                self.device,
                "logcat",
                "--format=tag",
                "--pid",  # This option is available since API level 24.
                pid,
            ]
            # Filter out some noisy and useless tags.
            + [f"{tag}:S" for tag in ["EGL_emulation"]]
            + (["--format=color"] if self.tools.input.is_color_enabled else []),
            env=self.tools.android_sdk.env,
            encoding="UTF-8",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
        )

    def logcat_tail(self, since: datetime):
        """Show the tail of the logs for Python-like apps, starting from a given
        timestamp.

        :param since: The start time from which logs should be displayed
        """
        try:
            # As best as we can make out, adb logcat returns UTF-8 output.
            # See #1425 for details.
            self.tools.subprocess.run(
                [
                    self.tools.android_sdk.adb_path,
                    "-s",
                    self.device,
                    "logcat",
                    "--format=tag",
                    "-t",
                    since.strftime("%m-%d %H:%M:%S.000000"),
                    "-s",
                    # This is a collection of log labels that should catch
                    # most Python app output.
                    "MainActivity:*",
                    "stdio:*",
                    "python.stdout:*",
                    "AndroidRuntime:*",
                ]
                + (["--format=color"] if self.tools.input.is_color_enabled else []),
                env=self.tools.android_sdk.env,
                check=True,
                encoding="UTF-8",
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError("Error starting ADB logcat.") from e

    def pidof(self, package: str, **kwargs) -> str | None:
        """Obtain the PID of a running app by package name.

        :param package: The package ID for the application (e.g.,
            ``org.beeware.tutorial``)
        :returns: The PID of the given app as a string, or None if it isn't
        running.
        """
        # The pidof command is available since API level 24. The level 23 emulator image also
        # includes it, but it doesn't work correctly (it returns all processes).
        try:
            # Exit status is unreliable: some devices (e.g. Nexus 4) return 0 even when no
            # process was found.
            return self.run("shell", "pidof", "-s", package, **kwargs).strip() or None
        except subprocess.CalledProcessError:
            return None

    def pid_exists(self, pid: str, **kwargs) -> bool:
        """Confirm if the PID exists on the emulator.

        :param pid: The PID to check
        :returns: True if the PID exists, False if it doesn't.
        """
        # Look for the existence of /proc/<PID> on the device filesystem.
        # If that file exists, so does the process.
        try:
            self.run("shell", "test", "-e", f"/proc/{pid}", **kwargs)
            return True
        except subprocess.CalledProcessError:
            return False

    def kill(self):
        """Stop the running Android emulator."""
        try:
            self.run("emu", "kill")
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError("Error stopping the Android emulator.") from e

    def datetime(self) -> datetime:
        """Obtain the device's current date/time.

        This date/time is naive (i.e. not timezone aware) and in the device's "local"
        time. Therefore, it may be quite different from the date/time for Briefcase and
        caution should be used if comparing it to machine's "local" time.
        """
        datetime_format = "%Y-%m-%d %H:%M:%S"
        try:
            device_datetime = self.run("shell", "date", f"+'{datetime_format}'").strip()
            return datetime.strptime(device_datetime, datetime_format)
        except (ValueError, subprocess.CalledProcessError) as e:
            raise BriefcaseCommandError("Error obtaining device date/time.") from e
